local SRB2CEversion = "1.1.1"

--[[ SRB2 Chaotic Edition
Author: birbhorse aka EeveeEuphoria/scorbuny
Additional credits: Mirror Mode/MirrorControls_v2.1.lua by QuickWolfPT
Additional credits: BootlegAttract by Tatsuru (found in the slot machines addon)
Additional credits: deepcopy by lua-users (http://lua-users.org/wiki/CopyTable)
Additional credits: customsave_io.lua (Serialize and Deserialize functions) by sprkizard AKA szv (found here: https://github.com/sprkizard/srb2-lua-lib/blob/master/game/customsave_io.lua)
]]


--[[ CHANGES COMPARED TO THE LAST VERSION:
    - Updated the icon used for "No Abilities" when not playing in amplify mode.
    - Also updated all the icons to have palettes that closer match the SRB2 color palette, making it so shading is more visible.
    - Made this mod easier to work w/ in regards to the "& Tails" bot found in single-player
        - A lot of effects that are either irrelevant or make it unreasonably unfair has been removed from being used by the bot
        - Text now displays showing when the 2nd player gets an effect
]]

--[[ NOTES TO SELF ABOUT POTENTIAL KART VERSION
    - use the icon i made for cement shoes as cement...tires? it's just CEICD0 kart.png
    - comment out the effects related to rings, srb2kart will have rings in a future update
        - once the next update comes out, which will merge with srb2 2.2, i have the potential to simplify the mod and make it compatible for both games
    - see if there's any useful code in that incomplete version i made awhile ago

]]

--[[ NEW EFFECT IDEAS
    - [ ] Rotater
        - Have sprite be rotated 180 degrees, making it appear wrong. Doesn't affect gameplay but it looks funny.
        - TEST WITH THE SIMPLIFIED CONTROL SCHEME
    - [ ] Resizer
        - Scales random objects around the player.
        - Unsure how to go about this w/o lag, but I'd like there to be a timer that decrements if the object has been scaled.
    - [ ] Always accelerating forward.
        - Not always pushing "walk forward", it literally propels you forward in the angle the character is facing.
        - Incompatile with Windy Day.

    - [ ] Press a certain button or get hurt
        - Try not to use custom buttons.
        - [ ] Ensure it works with "Remove Random Effect", otherwise it can get messy.
    - [ ] invisible hud
		- doesn't apply to chaos hud, unless in bad mode, in which case it'll slowly fade away
    - [ ] Simulated Lag
        - An effect of pure evil, don't apply it on netgames it'd only make it so much worse. ... unless >:)
        - Have a table with queued up button presses, in which they have been offset by a set amount of tics. And when the player presses a button it's ignored and put into there.
    - [ ] Randomize Abilities
        - In normal play, this randomzies your abilities for a set period of time to something else. Have it apply the stats of regular characters associated with said abilities.
        - In Bad Mode, this is permanently changed until a character swap, or another one of this.
    - [ ] Exploded Rings
        - [ ] Try adding that effect I've had for a long time, where rings scatter out of the player. Probably requires invincibility frames to prevent re-grabbing

    - [ ] Spinner
        - Player is visually spinning in place, but controls are unaffected

    - [?] Firework
        - [ ] Create a more polished explosion effect at the end

    - [ ] Try thinking of a more long-term effect, there's no restriction on the amount of effects now, should go wild.
        - [ ] Extend the amount of time Change Sprite goes for, once I fix that.

    - [?] Lucky!
        - [ ] Effect grants the player better luck for it's duration
            - [ ] Chance for certain effects to do other things
                - [ ] Have Spawn Object be nicer?
            - [x] Give twice as many rings on some occasions
            - [?] Random chance for a bad effect to be reduced by a second.
                - "if value.effect == "bad"" and random chance, then timer = $ - TICRATE"
        - [?] Have more visible effects happen besides just passive ones,
            - [ ] Random chance for a ring to pop out of the player, as if it spilled out of the player, with a ring sound playing too.
            - [x] Random chance for a nearby enemy to turn into a monitor, unless it's a boss.

    -- MISSING EFFECTS --

    - [ ] Change Sprite
        - Do this from the ground-up, used to cause a lot of SIGSEGVs, and I don't want that anymore.
        - [ ] Add more sprites
    - [ ] The Floor Is Lava
        - Should I even keep this one? Maybe just for amplify mode.
        - FRACUNIT/35, 5 seconds (10 in amplify)
    - [ ] Randomize SFX
        - Should also think about re-doing this, I don't think it's netplay safe still so oh well.
        - FRACUNIT/9, 20 seconds
]]

--[[ BUGS:
    - [ ] From last version:
        - characters like neo sonic also modify the camera, make it so the camera permanetly stays in the positiion it's set to be in

    - [ ] Sounds are weird in split-screen, player 2 has no volume adjustment, and player 1's sounds are halfed in volume? Sounds more like a bug with SRB2 itself.

    - [?] Do not add more roulettes if player is not alive, or is spectating.
        - [?] When the player dies, pause all roulettes and effects, and lower the HUD at a fast rate.

    - [?] Stop everything when player is disconnected and their body is left behind

    - [ ] Change Sprite doesn't remove itself properly when the player is a ring sprite in a minecart
    - [ ] Change Sprite ring still freezes Fang in place when shooting

    - [ ] When the special stage is ended, it uses a different way to say it's finished? Check for lines that require PF_FINISHED, and add whatever check that is.

    - [ ] Mods that use the anti-ring SFX seems to conflict with this; make it a seperate SFX file.

    - [ ] Changing settings also changes everyone else's settings? I dunno, make sure it doesn't!
]]

--[[ TO-DO:
    - [ ] Change the icon for "No Abilities" in non-amplify mode to be a spin-dash with a cross through it
        - Have it be the original icon in amplify mode

    - [ ] Have effects that don't decrement their timer, but instead decrement it based on how many times it's been used (Bouncy Floors comes to mind)

    - [ ] Have "Incline Probems" only work when the player is ascending (keep track of the player's last Z coordinate, then compare to the next one after)

    - [ ] Have chaos command work in the main menu, i.e. no player

    - [?] Allow amplify mode effects if bypassed using the effect button

    - [ ] Make ring monitors work with both 2x and negative rings
        - Need to do it via the ring monitor icon I think?
    - [ ] Also make these work for the ring weapons

    - [?] Add an option named "noincompat", disabling the incompatbility check. All hell will break loose.
        - [ ] Make it "on/off/max", in which max disables almost all checks, besides a couple.

    - [ ] Start using the "func" value more, and iron out any bugs with it.
        - Maybe include special values for functions n such that it'd need?

    - [ ] Fixed the problem with Change Color making teams not show up in match mode, maybe I should re-enable it for match mode? Dunno.

    - [ ] REWRITE "Change Sprite" ENTIRELY BY SCRATCH!
        - Code is a nightmare and it causes instability. I think it's responsible for the only sigsegvs seen in this, and it also fucks with Fang's ability.
        - [ ] Use a friggin table to manage everything.

    - [ ] Random scale for Spawn Object? That'd be kinda neat.
        - 33% chance to make a random scale, and if so, have it be between 50% and 150%

    - [?] Idea for net-games: Show player's current effects above their head. Would need to create an object for this, and figure out how to center it all.
        - [ ] IF USING PLAYERTHINK, USE A SEPERATE HOOK FOR IT
        - [ ] centering logic: do it like normal (i.e. for each index, add a certain amount), but take the total amount of indexes and do math to offset it
        - [ ] MAKE IT INVISIBLE TO ONLY THE PLAYER WHO HAS IT
            - In the object's thinker, check if [ consoleplayer ] is the player who shouldn't see it. If true, add [ MF2_DONTDRAW ], if false, remove

    - [?] Have this work in match mode, that'd be really cool to have.
        - [ ] Also make some exclusive effects for this, or tweak existing ones (i.e. "i taketh")
            - [ ] "i taketh" takes random ammo away
                - [ ] In Amplified mode, it also takes away a random weapon.
            - [ ] "Give Good Stuff" can give you a random weapon
        - [x] Make certain effects not work in match using "nomatch"
        - [?] Move up the HUD for match mode
            - [ ] Move the text up as well

    - [?] Put in "someone using blackhole" code
        - [x] variable: issomeoneusingblackhole

    - [ ] Add some compatibility to character mods and the like
        - Have "No Abilities" and "i taketh" work properly, that is the priority.

    - [?] Random chance for "drifting left" to turn into "drifting right"
        - Sprite: CEIDD0

    - [ ] Have "i taketh" remove a random entry (that's beyond the first) from chaos memory, a very subtle effect but one that's a bit annoying.

    - [?] Make new effects for the HUD
        - [x] When roulette starts, have it come up from the bottom of the screen
        - [ ] When new effect is applied, have it do a transparent bouncy effect, where it zooms out, showing it got applied

    - [?] ADD IN CUSTOMIZATION FROM THE LAST VERSION!
        - Don't really feel like making a new menu, and don't want to use MMT integration since that needs an overhaul as well.
        - [x] Have commands be simple. Use "chaos" and change stuff that way. Still have "chaoshelp" for in case.
            - [x] It will change settings in "player.chaoticsettings.xxx" variables.
            - [x] Have a HUD element show up, just like the last version of Chaos Edition, indicating you should read help.
            - OPTIONS:
                - [?] HUD Positioning
                    - [?] When the value is changed, move everything with it. i.e. move the value.ytime stuff.
        - [?] Create a save/load system
            - [ ] Create a system to allow this to be used on the title screen, by using lua i/o stuff.
            - [ ] Get local variables saved properly

    - [ ] Random chance for "Fling Player" to become "YEET" and play a yeet sound
    - [ ] Have "Fling Player" also work side ways in bad mode, or just allow for any angle, idk.

    - [ ] Add a "modding API" thing, this time do it right
        - Make sure to keep in mind that each player has their own copy of the effect table, it will need to be re-copied once any mods are loaded, so make a function for that

    - [ ] Add to the "Give Good Stuff" thing, like have tails' fly meter be extended or have the timer for that reset to 0, not sure how it works
        - [ ] Also reset the breath timer for underwater/space

    - [ ] Remove effects on player spawn if player is in boss, special, etc. due to the respawn stuff

    - [ ] NEW MECHANIC: In the last check function, have it so certain effects (i.e. cement shoes when no abilities is applied, or
        drifting left w/ the ride never ends) have a lesser chance of happening
        - Of course make it not applicable for Bad Mode.

    - [ ] Use some of the new music functions to detect current music, ensuring Random Music actually changes tracks.
        - [ ] Have "Random Music" show what new track is playing
]]

local issomeoneusingblackhole = false


freeslot(
"S_COLORRING",
"MT_JIMBO",
"S_CHAOSPIZZA",
"SPR_CHPI",
"sfx_chapis",
"sfx_chapki",
"sfx_boo",
"sfx_slappy",
"sfx_chgc",
"sfx_antico",
"S_CHAOSEXPLOSPIZZA",
"SPR_SOSS",
"S_SOSS_RUN1",
"S_SOSS_RUN2",
"S_SOSS_RUN3",
"S_SOSS_RUN4",
"S_SOSS_RUN5",
"S_SOSS_RUN6",
"MT_CHAOSROUL",
"S_CHAOSROUL",
"SPR_CEIB",
"SPR_CEIC",
"SPR_CEID",
"sfx_chadet",
"sfx_camin",
"sfx_camout"
)

states[S_SOSS_RUN1] = {
	sprite = SPR_SOSS,
    frame = A,
	tics = 3,
    nextstate = S_SOSS_RUN2
}

states[S_SOSS_RUN2] = {
	sprite = SPR_SOSS,
    frame = B,
	tics = 3,
    nextstate = S_SOSS_RUN3
}

states[S_SOSS_RUN3] = {
	sprite = SPR_SOSS,
    frame = C,
	tics = 3,
    nextstate = S_SOSS_RUN4
}

states[S_SOSS_RUN4] = {
	sprite = SPR_SOSS,
    frame = D,
	tics = 3,
    nextstate = S_SOSS_RUN5
}

states[S_SOSS_RUN5] = {
	sprite = SPR_SOSS,
    frame = E,
	tics = 3,
    nextstate = S_SOSS_RUN6
}

states[S_SOSS_RUN6] = {
	sprite = SPR_SOSS,
    frame = F,
	tics = 3,
    nextstate = S_SOSS_RUN1
}


states[S_CHAOSPIZZA] = {
	sprite = SPR_CHPI,
	tics = -1
}

states[S_CHAOSEXPLOSPIZZA] = {
	sprite = SPR_CHPI,
	frame = B,
	tics = -1
}

states[S_COLORRING] = {
	sprite = SPR_TRNG,
	frame = FF_ANIMATE|FF_GLOBALANIM,
	tics = -1,
	var1 = 23,
	var2 = 1,
	nextstate = S_COLORRING
}

mobjinfo[MT_JIMBO] = {
	doomednum = -1,
	spawnstate = S_CHAOSPIZZA,
	speed = 36*FRACUNIT,
	height = 128*FRACUNIT,
	radius = 25*FRACUNIT,
	flags = MF_NOGRAVITY|MF_NOCLIP
}

sfxinfo[sfx_chapis] = {
	singular = false,
	priority = 64,
	flags = SF_X4AWAYSOUND,
	caption = "Thud-um"
}

sfxinfo[sfx_chapki] = {
	singular = false,
	priority = 129,
	caption = "\"Oh shit.\""
}

sfxinfo[sfx_boo] = {
	singular = false,
	priority = 64,
	caption = "BoOoO!"
}

sfxinfo[sfx_slappy] = {
	singular = false,
	priority = 64,
	caption = "Slapped"
}

sfxinfo[sfx_chgc] = {
	singular = false,
	priority = 64,
	caption = "Coiled spring"
}

sfxinfo[sfx_chadet] = {
	singular = false,
	priority = 64,
	caption = "Flare gun"
}

mobjinfo[MT_CHAOSROUL] = {
	doomednum = -1,
	spawnstate = S_CHAOSROUL,
	height = 16*FRACUNIT,
    radius = 25*FRACUNIT,
	flags = MF_NOGRAVITY|MF_NOCLIP
}

states[S_CHAOSROUL] = {
	sprite = SPR_CEIC,
    frame = L,
	tics = -1,
}

local function Serialize(input)
	if type(input) == "string" then
		return "[[" .. input:gsub("](%--)]", "]-%1]") .. "]]"
	elseif type(input) == "number" then
		return "[" .. input .. "]"
	elseif type(input) == "boolean" then
		if input then
			return "T"
		else
			return "F"
		end
	elseif type(input) == "nil" then
		return "N"
	elseif type(input) == "table" then

		local str = "{"
		
		local delimit = false
		for k,v in pairs(input) do
			if delimit then
				str = $ .. ";"
			end
			
			str = $ .. Serialize(k) .. ":" .. Serialize(v)
			
			delimit = true
		end
		
		str = $ .. "}"
		return str
		
	else
		error("table contains a non-serializable element!")
	end
end

-- Converts the text representation of a Lua table back to the original table.
local function Deserialize(input)
	local output

	if input:sub(1, 2) == "[[" then -- string
		local endpos = input:find("]]")
		output = input:sub(3, endpos-1):gsub("]%-(%--)]", "]%1]")
		input = $:sub(endpos+2)
	elseif input:sub(1, 1) == "[" then -- number
		local endpos = input:find("]")
		output = tonumber(input:sub(2, endpos-1))
		input = $:sub(endpos+1)
	elseif input:sub(1, 1) == "T" then -- true
		output = true
		input = $:sub(2)
	elseif input:sub(1, 1) == "F" then -- false
		output = false
		input = $:sub(2)
	elseif input:sub(1, 1) == "N" then -- nil
		output = nil
		input = $:sub(2)
	elseif input:sub(1, 2) == "{}" then -- empty table
		output = {}
		input = $:sub(3)
	elseif input:sub(1, 1) == "{" then -- table
		output = {}
		
		while input:sub(1, 1) ~= "}" do
			local key, value
			
			key, input = Deserialize(input:sub(2))
			if input:sub(1, 1) ~= ":" then
				error("string is badly formatted! "..input)
			end
			
			value, input = Deserialize(input:sub(2))
			
			output[key] = value
		end
		
		input = $:sub(2)
	else
		error("string is badly formatted! "..input)
	end
	
	if input:len() then
		return output, input
	else
		return output
	end
end

local function DebugPrint(player, text)
    if player.chaoticsettings and player.chaoticsettings.debug == "on" then
        CONS_Printf(player, text)
    end
end

local function S_ChaoticSFX(player, sfx, value)
    if value.freeze then
        return false
    end

    if value.type == "Roulette" then
        return false
    end

    local sound
    if sfx == "start" then
        if value.startsfx then
            sound = value.startsfx
        end
    elseif sfx == "end" then
        if value.endsfx then
            sound = value.endsfx
        end
    end

    if sound == nil then
        DebugPrint(player, "Caution: No sound is playing for: " .. value.name .. " in " .. sfx)
        return false
    else
        if value.localsfx then
            S_StartSound(player.mo, sound, player)
        else
            S_StartSound(player.mo, sound)
        end
    end
end

local BtnToName = {
["off"] = off,
["custom1"] = BT_CUSTOM1,
["custom2"] = BT_CUSTOM2,
["custom3"] = BT_CUSTOM3,
["weaponnext"] = BT_WEAPONNEXT,
["weaponprev"] = BT_WEAPONPREV,
["ringtoss"] = BT_ATTACK,
["ringtossnormal"] = BT_FIRENORMAL,
["spin"] = BT_SPIN,
["jump"] = BT_JUMP,
["tossflag"] = BT_TOSSFLAG
}

local BtnList = { -- would be neat to make a function to automatically make this list lol
"off",
"custom1",
"custom2",
"custom3",
"weaponnext",
"weaponprev",
"ringtoss",
"ringtossnormal",
"spin",
"jump",
"tossflag"
}

local OptionsTable = {} -- used for options

local function addOption(tbl)
	table.insert(OptionsTable, tbl)
end

addOption{name = "save", Fname = "Save Settings", cat = "Options", playervar = true, optiontype = "save", noset = true, description = "Saves your current Chaotic Edition settings."}
addOption{name = "load", Fname = "Load Settings", cat = "Options", playervar = true, optiontype = "load", noset = true, description = "Loads in your Chaotic Edition settings."}
addOption{name = "amplify", Fname = "Amplify Mode", cat = "Options", options = {"on", "off"}, set = "off", admin = true, description = "Makes most effects more intense, and introduces new exclusive effects."}
addOption{name = "limit", Fname = "Effect Limit", cat = "Options", options = {0, 8}, playervar = true, set = 0, optiontype = "number", description = "Limits the amount of effects that can be applied to the player."}
addOption{name = "carryover", Fname = "Carry Over Effects", cat = "Options",  options = {"on", "off"}, playervar = true, set = "on", description = "Carries effects between respawns and level loads."}
addOption{name = "memory", Fname = "Memory List", cat = "Options", options = {0, 6}, playervar = true, set = 5, optiontype = "number", description = "Allows a certain amount of effects to be remembered, preventing them from occuring until the item is removed from the list."}
addOption{name = "position", Fname = "HUD Position", cat = "HUD", options = {"top", "bottom"}, playervar = true, set = "bottom", description = "Adjusts the position of the HUD."}
addOption{name = "timerbar", Fname = "Timer Bar", cat = "HUD",  options = {"on", "off"}, playervar = true, set = "on", description = "Shows a bar displaying the time until the next effect."}
addOption{name = "rainbow", Fname = "Rainbow Effect", cat = "HUD",  options = {"on", "off"}, playervar = true, set = "off", description = "Enables the rainbow effect on the timer bar."}
addOption{name = "afterfinish", Fname = "Continue After Finished", cat = "Tweaks", options = {"on", "off"}, playervar = true, set = "off", description = "Allows effects to continue to be applied after completeing a level, useful for netplay."}
addOption{name = "minwait", Fname = "Minimum Wait Time", cat = "Tweaks", options = {3, 8}, optiontype = "number", playervar = true, set = 4, description = "Changes the minimum amount of time until a new effect occurs."}
addOption{name = "maxwait", Fname = "Maximum Wait Time", cat = "Tweaks", options = {4, 16}, optiontype = "number", playervar = true, set = 12, description = "Changes the maximum amount of time until a new effect occurs."}
addOption{name = "roulettesfx", Fname = "Roulette Sounds", cat = "Tweaks", options = {"on", "off"}, playervar = true, set = "on", description = "Lets the effect roulette play sounds as it spins through effects."}
addOption{name = "debug", Fname = "Debugging", cat = "Misc.", options = {"on", "off"}, playervar = true, set = "off", description = "Allows for debug functionality, such as printing text to console."}
addOption{name = "noincompat", nomatch = true, Fname = "No Incompatbility Check", cat = "Misc.", options = {"on", "off"}, set = "off", description = "Disables the incompatbility check for all effects. All hell will break loose."}
addOption{name = "autosave", Fname = "Auto-save settings", cat = "Misc.", options = {"on", "off"}, playervar = true, set = "on", description = "Automatically saves settings when a setting is changed."}
addOption{name = "autoload", Fname = "Auto-load settings", cat = "Misc.", options = {"on", "off"}, playervar = true, set = "on", description = "Automatically loads settings when the player first spawns."}
addOption{name = "effectbtn", nomatch = true, Fname = "Effect Button", cat = "Misc.", options = BtnList, seperateoptions = true, playervar = true, set = "off", description = "Instantly applies an effect to the player when pressed. Not available in competitve modes."}

addOption{name = "ReadHelp", Fname = "YOU SHOULD NOT BE SEEING THIS", hide = true, options = {true, false}, playervar = true, set = false, description = "This is a janky thing, you shouldn't see this, but I'm not surprised you can see this."}

-- Options, Tweaks, Misc.

local function RandomLetter()
    local NUM = P_RandomRange(1, 26)
    if NUM == 1 then
        return "a"
    elseif NUM == 2 then
        return "b"
    elseif NUM == 3 then
        return "c"
    elseif NUM == 4 then
        return "d"
    elseif NUM == 5 then
        return "e"
    elseif NUM == 6 then
        return "f"
    elseif NUM == 7 then
        return "g"
    elseif NUM == 8 then
        return "h"
    elseif NUM == 9 then
        return "i"
    elseif NUM == 10 then
        return "j"
    elseif NUM == 11 then
        return "k"
    elseif NUM == 12 then
        return "l"
    elseif NUM == 13 then
        return "m"
    elseif NUM == 14 then
        return "n"
    elseif NUM == 15 then
        return "o"
    elseif NUM == 16 then
        return "p"
    elseif NUM == 17 then
        return "q"
    elseif NUM == 18 then
        return "r"
    elseif NUM == 19 then
        return "s"
    elseif NUM == 20 then
        return "t"
    elseif NUM == 21 then
        return "u"
    elseif NUM == 22 then
        return "v"
    elseif NUM == 23 then
        return "w"
    elseif NUM == 24 then
        return "x"
    elseif NUM == 25 then
        return "y"
    elseif NUM == 26 then
        return "z"
    end
end

local function RandomWord()
    local WORD = RandomLetter() .. RandomLetter() .. RandomLetter().. RandomLetter().. RandomLetter().. RandomLetter().. RandomLetter().. RandomLetter().. RandomLetter().. RandomLetter().. RandomLetter().. RandomLetter()
    return WORD
end


local function tablelength(T)
    local count = 0
    for _ in pairs(T) do count = count + 1 end
    return count
end

local function searchOption(name)
    for index, value in ipairs(OptionsTable) do
        if name == value.name then
            return value.set
        end
    end
end

local EffectsTable = {}

local vibechance = FRACUNIT -- (FRACUNIT/2) + (FRACUNIT/3)

local amplifymode = "off"

local function addEffect(tbl)
	table.insert(EffectsTable, tbl)
end

local function RemoveEffect(player, index, value, nomem)
    local name = value.name
    S_ChaoticSFX(player, "end", value)
    if index and player.chaoticeffects[index].object and player.chaoticeffects[index].object.valid then
        P_RemoveMobj(player.chaoticeffects[index].object)
    end
    table.remove(player.chaoticeffects, index)
    if name == "Slapped" then -- restore original name
        name = "Fling Player"
    end

    if not nomem and player.chaoticsettings.memory > 0 then -- add to the player.chaoticmemory table, and remove first entry if table is too full
        table.insert(player.chaoticmemory, name)
        if tablelength(player.chaoticmemory) > tonumber(player.chaoticsettings.memory) then
            table.remove(player.chaoticmemory, 1)
        end
    end
end

  addEffect{name = "Roulette", chance = 1/FRACUNIT, duration = 0, sprite = "CEICB0", type = "Roulette", nohudtext = true, notimer = true} -- hacky way of making a new roulette system
  addEffect{name = "Cement Shoes", chance = FRACUNIT/9, duration = 7, sprite = "CEICD0", addition = 1, nocart = true, effect = "bad", nonights = true,
startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true,
incompatible = {"Incline Problems", "I WAS FROZEN TODAY!", "Change Color", "Butter Physics", "Black Hole", "Bouncy Floors", "Gravity Coil", "Super Jump", "Moonjump", "Zero Gravity"}} -- so many items
  addEffect{name = "No Abilities", chance = FRACUNIT/12, duration = 13, sprite = "CEIDL0", effect = "bad", addition = 2, nocart = true,
startsfx = sfx_kc59, endsfx = sfx_kc65,
incompatible = {"Moonjump"}}
  addEffect{name = "Funky Camera", chance = FRACUNIT/23, duration = 10, sprite = "CEICZ0", effect = "bad", addition = 2, nocart = true, nonights = true, nobot = true,
localsfx = true} -- startsfx = sfx_kc59, endsfx = sfx_kc65,
  addEffect{name = "Giant", chance = FRACUNIT/9, duration = 8, sprite = "CEICC0", nonights = true, nocart = true, nospecialstage = true, nospawn = true, effect = "bad",
startsfx = sfx_kc5a, endsfx = sfx_kc59,
incompatible = {"Tiny"}}
  addEffect{name = "Tiny", chance = FRACUNIT/10, duration = 10, sprite = "CEICG0", addition = 2, nocart = true, nospawn = true, effect = "bad", nonights = true,
startsfx = sfx_kc59, endsfx = sfx_kc5a,
incompatible = {"Giant"}}
  addEffect{name = "Fling Player", chance = FRACUNIT/8, duration = 0, sprite = "CEICB0", nocart = true, nohudtext = true, effect = "bad", nonights = true,
incompatible = {"I WAS FROZEN TODAY!"}}
  addEffect{name = "Random Music", chance = FRACUNIT/14, duration = 0, sprite = "CEIC00", effect = "neutral", nobot = true}
  addEffect{name = "I WAS FROZEN TODAY!", chance = FRACUNIT/26, duration = 3, sprite = "CEIC60", addition = 1, nospawn = true, effect = "bad", nocart = true,
startsfx = sfx_s3k80, endsfx = sfx_s3k80,
incompatible = {"Change Color", "Cement Shoes", "Change Sprite", "Black Hole"}}
  addEffect{name = "Spawn Object", chance = FRACUNIT/18, duration = 0, sprite = "CEIC80", nonights = true, nohudtext = true, effect = "neutral",
incompatible = {"I WAS FROZEN TODAY!"}}
  addEffect{name = "Butter Physics", chance = FRACUNIT/11, duration = 14, sprite = "CEICI0", addition = 2, nonights = true, nocart = true, effect = "bad",
startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true,
incompatible = {"Cement Shoes", "Incline Problems"}}
  addEffect{name = "Drifting Left", chance = FRACUNIT/25, duration = 10, sprite = "CEICH0", addition = 1, effect = "bad", nocart = true, nonights = true, nohudtext = true, nobot = true,
startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true,
incompatible = {"No Turning"}}
  addEffect{name = "Negative Rings", chance = FRACUNIT/16, duration = 12, sprite = "CEIC20", addition = 5, noboss = true, nospawn = true, effect = "bad", nobot = true,
startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true,
incompatible = {"2x Rings", "Bleeding"}}
  addEffect{name = "Change Character", chance = FRACUNIT/17, duration = 4, sprite = "CEICK0", effect = "neutral", endsfx = sfx_kc65,
incompatible = {"I WAS FROZEN TODAY!"}}
  addEffect{name = "Change Color", chance = FRACUNIT/10, duration = 3, sprite = "CEICJ0", effect = "neutral", nomatch = true, endsfx = sfx_kc65,
incompatible = {"I WAS FROZEN TODAY!", "Cement Shoes"}}
  addEffect{name = "the ride never ends", chance = FRACUNIT/25, duration = 15, sprite = "CEICB0", addition = 4, effect = "bad", nonights = true, nospawn = true,
startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true}
  addEffect{name = "Nuke", chance = FRACUNIT/40, duration = 0, sprite = "CEICS0", nonights = true, effect = "good", nospecialstage = true}
  addEffect{name = "Darkened Vision", chance = FRACUNIT/23, duration = 15, sprite = "CEIC30", effect = "bad", addition = 5, nobot = true,
startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true}
  addEffect{name = "Invert Controls", chance = FRACUNIT/45, duration = 8, sprite = "CEICX0", effect = "bad", badmode = true,
startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true,
incompatible = {"I WAS FROZEN TODAY!", "Laggy Controls", "Busted Controller"}}
  addEffect{name = "i taketh", chance = FRACUNIT/25, duration = 0, sprite = "CEICX0", effect = "bad"}
  addEffect{name = "Remove Random Effect", chance = FRACUNIT/22, duration = 0, sprite = "CEICO0", effect = "neutral"}
  addEffect{name = "Gravity Coil", chance = FRACUNIT/18, duration = 15, sprite = "CEICV0", addition = 3, nonights = true, effect = "neutral", nocart = true,
startsfx = sfx_chgc, endsfx = sfx_kc65, localsfx = true,
incomaptible = {"Super Jump", "Bouncy Floors"}}
  addEffect{name = "Super Jump", chance = FRACUNIT/30, duration = 12, sprite = "CEICL0", addition = 1, nocart = true, nonights = true, effect = "neutral",
startsfx = sfx_kc5a, endsfx = sfx_kc59, localsfx = true,
incompatible = {"Gravity Coil", "Bouncy Floors", "Moonjump"}}
  addEffect{name = "the pizza is aggressive", chance = FRACUNIT/37, duration = 10, sprite = "CEICM0", addition = 3, effect = "bad", nonights = true, nospawn = true, nospecialstage = true, noboss = true,
startsfx = sfx_chapis, endsfx = sfx_kc65}
  addEffect{name = "Ghost", chance = FRACUNIT/25, duration = 8, sprite = "CEICT0", effect = "bad", nospawn = true,
startsfx = sfx_boo, endsfx = sfx_kc65}
--  addEffect{name = "Change Sprite", chance = FRACUNIT/10, duration = 18, sprite = "CEICY0", nospawn = true,
--startsfx = sfx_kc65, endsfx = sfx_kc65,
--incompatible = {"I WAS FROZEN TODAY!"}}
  addEffect{name = "Give Good Stuff", chance = FRACUNIT/24, duration = 0, effect = "good", sprite = "CEICB0"}
  addEffect{name = "Earthquake", chance = FRACUNIT/19, duration = 10, effect = "bad", sprite = "CEICU0", nomatch = true,
startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true}
  addEffect{name = "Money Bag", chance = FRACUNIT/31, duration = 4, effect = "good", sprite = "CEICN0", nobot = true,
startsfx = sfx_s3kb3, endsfx = sfx_chchng,}
  addEffect{name = "Superring", chance = FRACUNIT/15, duration = 0, effect = "good", sprite = "CEICP0", nobot = true}
  addEffect{name = "Anti-Superring", chance = FRACUNIT/16, duration = 0, effect = "bad", sprite = "CEICQ0", noboss = true, nobot = true}
  addEffect{name = "COMBO TIME!", chance = FRACUNIT/20, duration = -1, effect = "neutral", sprite = "CEIDA0", notimer = true, nobot = true,
startsfx = sfx_kc48}
  addEffect{name = "Reversed!", chance = FRACUNIT/34, duration = 0, effect = "bad", sprite = "CEICH0", badmode = true, nocart = true, nonights = true}
  addEffect{name = "Bouncy Floors", chance = FRACUNIT/24, duration = 7, effect = "bad", sprite = "CEIDB0", addition = 2, nocart = true, nonights = true,
  startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true,
incompatible = {"Super Jump", "Gravity Coil", "Cement Shoes", "Butter Physics", "Zero Gravity"}}
  addEffect{name = "Black Hole", chance = FRACUNIT/20, duration = 13, sprite = "CEIC50", addition = 2, effect = "bad", nospawn = true, nobot = true, nospecialstage = true, nonights = true, nocart = true, nomatch = true,
startsfx = sfx_s3k74, endsfx = sfx_kc65,
incompatible = {"Magnet", "Cement Shoes"}}
  addEffect{name = "Lucky!", chance = FRACUNIT/18, effect = "good", duration = 15, sprite = "CEIDC0", addition = 1,
  startsfx = sfx_s1c9, endsfx = sfx_kc65, localsfx = true}
  addEffect{name = "No Turning", chance = FRACUNIT/33, effect = "bad", duration = 7, sprite = "CEIC40", addition = 2, nocart = true, nonights = true, nobot = true,
startsfx = sfx_s3kb2, endsfx = sfx_kc65, localsfx = true,
incompatible = {"Drifting Left", "Busted Controller"}}
  addEffect{name = "Firework", chance = FRACUNIT/21, effect = "neutral", duration = 1, sprite = "CEIDE0", addition = 1, nocart = true, nospawn = true,
startsfx = sfx_chadet,
incompatible = {"Cement Shoes", "Super Jump", "Zero Gravity"}}
  addEffect{name = "Moonjump", chance = FRACUNIT/13, effect = "good", duration = 10, sprite = "CEIDF0", addition = 2, nocart = true, nonights = true, nobot = true,
startsfx = sfx_kc5a, endsfx = sfx_kc65, localsfx = true,
incompatible = {"Cement Shoes", "Super Jump", "No Abilities", "Zero Gravity"}}
  addEffect{name = "Zero Gravity", chance = FRACUNIT/28, effect = "bad", duration = 5, sprite = "CEIDG0", addition = 3, nocart = true, nonights = true,
startsfx = sfx_kc65, endsfx = sfx_kc65, localsfx = true,
incompatible = {"Cement Shoes", "Super Jump", "Moon Jump", "Bouncy Floors"}}
  addEffect{name = "Busted Controller", chance = FRACUNIT/20, effect = "bad", duration = 8, sprite = "CEIC40", nocart = true, badmode = true,
startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true,
incompatible = {"No Turning", "I WAS FROZEN TODAY!", "Laggy Controls"}}
  addEffect{name = "2x Rings", chance = FRACUNIT/13, duration = 10, sprite = "CEIDH0", addition = 5, noboss = true, nospawn = true, effect = "good", nohudtext = true, nobot = true,
startsfx = sfx_kc5a, endsfx = sfx_kc65, localsfx = true,
incompatible = {"Negative Rings"}}

  addEffect{name = "Windy Day", chance = FRACUNIT/19, duration = 8, sprite = "CEIDI0", addition = 3, effect = "bad", nonights = true,
startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true,
incompatible = {"Rotater"}}
  addEffect{name = "Incline Problems", chance = FRACUNIT/16, duration = 13, sprite = "CEIDJ0", addition = 5, effect = "bad", nonights = true, nospecialstage = true,
startsfx = sfx_kc59, endsfx = sfx_kc5a, localsfx = true,
incompatible = {"Butter Physics", "Cement Shoes"}}
  addEffect{name = "Bleeding", chance = FRACUNIT/10, duration = 12, sprite = "CEIDK0", addition = 7, effect = "bad", nonights = true, nobot = true,
startsfx = sfx_kc59, endsfx = sfx_kc5a, localsfx = true,
incompatible = {"Negative Rings"}}

--[[  addEffect{name = "Laggy Controls", chance = FRACUNIT/26, effect = "bad", duration = 8, sprite = "CEICQ0", badmode = true, nonet = true,
endsfx = sfx_kc65,
incompatible = {"Invert Controls"}}]]

--  addEffect{name = "TEST!", chance = FRACUNIT/99, effect = "good", duration = 10, sprite = "CEIDC0", addition = 1, func = TestFunc}

--table.insert(EffectsTable, tbl)

local EffectNames = {}
local EffectNames2 = {"off"}

for index, value in ipairs(EffectsTable) do
    EffectNames[value.name] = true
end

for index, value in ipairs(EffectsTable) do
    if value.name != "Roulette" then
        table.insert(EffectNames2, value.name)
    end
end

addOption{name = "blacklist", Fname = "Blacklist", cat = "Options", playervar = true, optiontype = "table", set = {}, options = EffectNames, hideoptions = true, description = "Adds/removes effects from a blacklist. Use \"add\" or \"remove\", plus an item, to change the list."}
addOption{name = "effecttype", nomatch = true, Fname = "Effect Type", cat = "Misc.", options = EffectNames2, seperateoptions = true, playervar = true, set = "off", description = "Supplements effectbtn, wherein a specific effect will be applied when the button is pressed. CASE SENSITIVE!"}

local ObjectsTable = {} -- used for "Spawn Object" effect

local function addObject(tbl)
	table.insert(ObjectsTable, tbl)
end

addObject{id = MT_RING_BOX, name = "Ring Box"}
addObject{id = MT_BLUECRAWLA, name = "Blue Crawla"}
addObject{id = MT_GOLDBUZZ, name = "Gold Buzz"}
addObject{id = MT_JETTBOMBER, name = "Jetty-Syn Bomber"}
addObject{id = MT_DETON, name = "Deton"}
addObject{id = MT_SPINCUSHION, name = "Spincushion"}
addObject{id = MT_FACESTABBER, name = "Lance-a-Bot"}
addObject{id = MT_BLUECRAWLA, name = "Blue Crawla"}
addObject{id = MT_EGGMOBILE, name = "Eggman"}
addObject{id = MT_GOOMBA, name = "Goomba"}
addObject{id = MT_REDSPRING, name = "Red Spring"}
addObject{id = MT_BLUESPRING, name = "Blue Spring"}
addObject{id = MT_TNTBARREL, name = "TNT Barrel"}
addObject{id = MT_PYREFLY, name = "Pyre Fly"}
addObject{id = MT_PENGUINATOR, name = "Penguinator"}
addObject{id = MT_ROBOHOOD, name = "Robo-Hood"}
addObject{id = MT_GFZTREE, name = "Tree"}
addObject{id = MT_PITY_BOX, name = "Pity Shield Box"}
addObject{id = MT_TURRET, name = "Turret"}

local function addHUDText(player, value)
    local tbl = {text = value, trans = 0, timer = 0, expire = (3*TICRATE) - 2}
    local sendplayer = player
    if player.bot > 0 then
        sendplayer = players[0]
        local prevtext = tbl.text
        tbl.text = "(\x88" .. "2P" .. "\x80) " .. prevtext
        tbl.timer = -2 * TICRATE -- makes the text display for longer
    end
    
    table.insert(sendplayer.chaotictext, tbl)
end

local function numberinserver() -- thanks to "Monster ɪestyn" on the SRB2 discord for this
    local c = 0
    for p in players.iterate do c = $ + 1 end
    return c
end

local function GetHelp(name, ind)
    for index, value in ipairs(OptionsTable) do
        if name == value.name or ind == index and not value.hide then
            return true
        end
    end
end

local function MusicRand(num) -- ctrack = current track
	if num == "RANGE" then
		return 70
	end

	if num == 1 then
		return "TUTORL"
	elseif num == 2 then
		return "GFZ1"
	elseif num == 3 then
		return "GFZ2"
	elseif num == 4 then
		return "VSBOSS"
	elseif num == 5 then
		return "VSALT"
	elseif num == 6 then
		return "THZ1"
	elseif num == 7 then
		return "THZ2"
	elseif num == 8 then
		return "DSZ1"
	elseif num == 9 then
		return "DSZ2"
	elseif num == 10 then
		return "CEZ1"
	elseif num == 11 then
		return "CEZ2"
	elseif num == 12 then
		return "ACZ1"
	elseif num == 13 then
		return "ACZ2"
	elseif num == 14 then
		return "VSFANG"
	elseif num == 15 then
		return "RVZ1"
	elseif num == 16 then
		return "RVZ2"
	elseif num == 17 then
		return "RVZALT"
	elseif num == 18 then
		return "DCZ"
	elseif num == 19 then
		return "ERZ1"
	elseif num == 20 then
		return "ERZ2"
	elseif num == 21 then
		return "DISCO"
	elseif num == 22 then
		return "BCZ1"
	elseif num == 23 then
		return "VSMETL"
	elseif num == 24 then
		return "VSBRAK"
	elseif num == 25 then
		return "FHZ"
	elseif num == 26 then
		return "MARIO1"
	elseif num == 27 then
		return "MARIO2"
	elseif num == 28 then
		return "THZALT"
	elseif num == 29 then
		return "HAUNTD"
	elseif num == 30 then
		return "AGZ"
	elseif num == 31 then
		return "ATZ"
	elseif num == 32 then
		return "CCZ"
	elseif num == 33 then
		return "DHZ"
	elseif num == 34 then
		return "APZ1"
	elseif num == 35 then
		return "APZ2"
	elseif num == 36 then
		return "SPEC1"
	elseif num == 37 then
		return "SPEC2"
	elseif num == 38 then
		return "SPEC3"
	elseif num == 39 then
		return "SPEC4"
	elseif num == 40 then
		return "SPEC5"
	elseif num == 41 then
		return "SPEC6"
	elseif num == 42 then
		return "SPEC7"
	elseif num == 43 then
		return "BHZ"
	elseif num == 44 then
		return "SPEC8"
	elseif num == 45 then
		return "OSPEC"
	elseif num == 46 then
		return "OSPEC2"
	elseif num == 47 then
		return "GFZALT"
	elseif num == 48 then
		return "AGZALT"
	elseif num == 49 then
		return "VSAGZ"
	elseif num == 50 then
		return "MP_WTR"
	elseif num == 51 then
		return "MP_ICE"
	elseif num == 52 then
		return "MP_RAI"
	elseif num == 53 then
		return "MP_AIR"
	elseif num == 54 then
		return "MP_SUN"
	elseif num == 55 then
		return "MP_DES"
	elseif num == 56 then
		return "MP_DOM"
	elseif num == 57 then
		return "MP_GHZ"
	elseif num == 58 then
		return "MP_MED"
	elseif num == 59 then
		return "MP_0RI"
	elseif num == 60 then
		return "MP_COA"
	elseif num == 61 then
		return "MP_FIR"
	elseif num == 62 then
		return "MP_FOR"
	elseif num == 63 then
		return "MP_FR2"
	elseif num == 64 then
		return "MP_GL2"
	elseif num == 65 then
		return "MP_GLZ"
	elseif num == 66 then
		return "MP_MIN"
	elseif num == 67 then
		return "MP_MRZ"
	elseif num == 68 then
		return "MP_WT2"
	elseif num == 69  then -- nice
		return "_CHSEL"
	elseif num == 70 then
		return "_INTER"

	end
end

local function numberofskins() -- modified on code from Monster ɪestyn
	local c = 0
	for p in skins.iterate do c = $ + 1 end
	return (c-1)
end

-- thanks to lua-users for the code! source: http://lua-users.org/wiki/CopyTable

local function deepcopy(orig) -- used solely to prevent a super weird glitch where the original effects table is affected by non-direct variable changes
    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[deepcopy(orig_key)] = deepcopy(orig_value)
        end
        setmetatable(copy, deepcopy(getmetatable(orig)))
    else -- number, string, boolean, etc
        copy = orig
    end
    return copy
end

local function ValidBHObject(mobj)
    if (mobj.flags & (MF_SCENERY|MF_SOLID)) == (MF_SCENERY|MF_SOLID) and not (mobj.type != MT_SPIKE and amplifymode == "on") then
        return false -- exception for stuff like GFZ trees, too annoying to deal with
    elseif (mobj.flags & MF_SPRING) and amplifymode == "off" then
        return false
    elseif mobj.type == MT_PLAYER then
        return false
	elseif (mobj.flags & MF_PUSHABLE) then 
        return false  -- just because those gargoyles are annoying too 
    elseif (mobj.type == MT_CHAOSROUL or mobj.type == MT_EGGTRAP or mobj.type == MT_STEAM or mobj.type == MT_CHAINPOINT or mobj.type == MT_MINECART or
    mobj.type == MT_MINECARTSEG or mobj.type == MT_MINECARTSPAWNER or mobj.type == MT_MINECARTEND or mobj.type == MT_MINECARTENDSOLID or
    mobj.type == MT_MINECARTSIDEMARK or mobj.type == MT_MINECARTSPARK) then
        return false
    else
        return true
    end
end

local function ValidLuckyObject(mobj)
   --[[ if (mobj.flags & (MF_SCENERY|MF_SOLID)) == (MF_SCENERY|MF_SOLID) and not (mobj.type != MT_SPIKE and amplifymode == "on") then
        return false -- exception for stuff like GFZ trees, too annoying to deal with
    elseif (mobj.flags & MF_SPRING) then
        return false
    elseif mobj.type == MT_PLAYER then
        return false
	elseif (mobj.flags & MF_PUSHABLE) or (mobj.flags & MF_MONITOR) or (mobj.flags & MF_BOSS) then
        return false
    elseif (mobj.type == MT_CHAOSROUL or mobj.type == MT_EGGTRAP or mobj.type == MT_STARPOST or mobj.type == MT_STEAM or mobj.type == MT_CHAINPOINT or mobj.type == MT_MINECART or
    mobj.type == MT_MINECARTSEG or mobj.type == MT_MINECARTSPAWNER or mobj.type == MT_MINECARTEND or mobj.type == MT_MINECARTENDSOLID or
    mobj.type == MT_MINECARTSIDEMARK or mobj.type == MT_MINECARTSPARK or mobj.type == MT_SIGN or mobj.type == MT_TOKEN) then
        return false]]
    if (mobj.flags & MF_ENEMY) and not (mobj.flags & MF_BOSS) then
        return true
    else
        return false
    end
end

local function IsMatchMode()
    local header = mapheaderinfo[gamemap].typeoflevel
    if (header & TOL_MATCH) or (header & TOL_TAG) or (header & TOL_CTF) then
        return true
    end

    return false
end

local function BootlegAttract(mo, target, speed, xoffs, yoffs, zoffs) -- originated from tatsuru's conversion of SRB2 source code
	xoffs = $ or 0
	yoffs = $ or 0
	zoffs = $ or 0
	
	local newx = target.x + xoffs
	local newy = target.y + yoffs
	local newz = target.z + zoffs
	local dist = R_PointToDist2(0, 0, R_PointToDist2(mo.x, mo.y, newx, newy), newz - mo.z)
	
	if speed < dist
		mo.momx = FixedMul(FixedDiv(newx - mo.x, dist), speed)
		mo.momy = FixedMul(FixedDiv(newy - mo.y, dist), speed)
		mo.momz = FixedMul(FixedDiv(newz - mo.z, dist), speed)
	elseif speed > dist
		mo.momx = 0
		mo.momy = 0
		mo.momz = 0
		P_TeleportMove(mo, newx, newy, newz)
	end
end

local numberofeffects = tablelength(EffectsTable)

local function GivePizza(player)
    player.chaotic.PizzaRemoval = nil
        local thepizzais = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.ceilingz-(6*FRACUNIT), MT_JIMBO)
        thepizzais.target = player.mo
 
        local size
        local choosesize = P_RandomRange(1, 4)
 
        if amplifymode == "on" then
            choosesize = P_RandomRange(2, 5)
        end
 
        if choosesize == 1 then
            size = "small"
        elseif choosesize == 2 then
            size = "medium"
        elseif choosesize == 3 then
            if P_RandomChance(FRACUNIT/2) then
                size = "large"
            else
                if P_RandomChance(FRACUNIT/2) then
                    size = "medium"
                else
                    size = "small"
                end
            end
        elseif choosesize == 4 then
            if P_RandomChance(FRACUNIT/99) then -- extrodionarly rare chance!
                size = "extra large"
            else
                if P_RandomChance(FRACUNIT/2 + FRACUNIT/3) then
                    size = "medium"
                else
                    if P_RandomChance(FRACUNIT/2) then
                        size = "small"
                    else
                        size = "large"
                    end
                end
            end
        elseif choosesize == 5 then -- only happens in bad mode
            size = "extra large"
        end

        if not size then -- failsafe, just for in case of some really weird situation occuring
            size = "small"
        end

        if size == "large" or size == "medium" or size == "small" or size == "personal" or size == "xl" or size == "extra large"  or size == "extralarge" then -- REMOVE THIS LATER
            thepizzais.sizeoption = size
        end

        local toppings = "pepperoni"

        if P_RandomChance(FRACUNIT/4) then
            thepizzais.state = S_CHAOSEXPLOSPIZZA
            toppings = "deton"
        end

        print(player.name .. " has ordered a " .. size .. " sized pizza, with " .. toppings .. " toppings to themselves.")
        if netgame then chatprint(player.name .. " has ordered a " .. size .. " sized pizza, with " .. toppings .. " toppings to themselves.", true) end
        player.chaotic.PizzaSpawned = true
end

-- !! ALSO APPLY ANY CHANGES HERE TO "PseudoTable" I'M TOO LAZY TO SEPERATE THE CODE CORRECTLY !!
local function NonConflictingEffect(player, override, neweffecttbl) 
-- !! ALSO APPLY ANY CHANGES HERE TO "PseudoTable" I'M TOO LAZY TO SEPERATE THE CODE CORRECTLY !!
    local tbl = player.chaoticeffects
    if not neweffecttbl then
        neweffecttbl = player.chaotic.StatusTable
    end
    local chaosmemory = player.chaoticmemory
    local incompatible = neweffecttbl.incompatible
    local nonights = neweffecttbl.nonights
    local waittime = 12

    if override then
        incompatible = nil
        nonights = nil
        chaosmemory = nil
        waittime = 0
    end

    if not IsMatchMode() and player.chaoticsettings.noincompat == "on" then
        incompatible = nil
    end

    --[[if player.chaotic.RouletteTimer < waittime then
        DebugPrint(player, "roulette timer isn't finished")
        return false
    end]]

    if neweffecttbl.type == "Roulette" then
        DebugPrint(player, "this is the roulette! free re-roll i guess, you have a 1/65536 to land on this lol")
        return false
    end

    if not override and (neweffecttbl.badmode and amplifymode != "on") then
        DebugPrint(player, "did not choose due to not being in bad mode")
        return false
    end

    if neweffecttbl.nonet and netgame then
        DebugPrint(player, "did not choose due to being in a netgame")
        return false
    end

    if neweffecttbl.nobot and player.bot > 0 then
        DebugPrint(player, "did not choose due to being tailsbot")
        return false
    end

    if neweffecttbl.nomatch and IsMatchMode() then
        DebugPrint(player, "did not choose due to being in match mode")
        return false
    end

    if neweffecttbl.name == "Remove Random Effect" and tablelength(tbl) < 2 then
        DebugPrint(player, "did not choose due to there not being enough effects")
        return false
    end
    if neweffecttbl.name == "Earthquake" and numberinserver() > 2 then
        DebugPrint(player, "did not choose due to 3+ players")
        return false
    end
    if (neweffecttbl.name == "Anti-Superring" or neweffecttbl.name == "Firework") and player.rings == 0 then
        DebugPrint(player, "did not choose due no rings")
        return false
    end
    if (mapheaderinfo[gamemap].typeoflevel & TOL_NIGHTS) and nonights then
        DebugPrint(player, "did not choose due to being in NiGHTs")
        return false
    end
    if neweffecttbl.nocart and (player.powers[pw_carry] == CR_MINECART or player.powers[pw_carry] == CR_ROLLOUT or player.powers[pw_carry] == CR_ZOOMTUBE) then
        DebugPrint(player, "did not choose due to being in cart/rollout/zoomtube")
        return false
    end
    if neweffecttbl.nospecialstage and G_IsSpecialStage(gamemap) then
        DebugPrint(player, "did not choose due to special stage")
        return false
    end
    if neweffecttbl.noboss and (mapheaderinfo[gamemap].bonustype == 1 or mapheaderinfo[gamemap].bonustype == 2) then
        DebugPrint(player, "did not choose due to being in a boss stage")
        return false
    end

    if player.chaoticsettings.blacklist then
        for key, value in pairs(player.chaoticsettings.blacklist) do
            if neweffecttbl.name == key then
                DebugPrint(player, "did not choose due being in blacklist")
                return false
            end
        end
    end
    
    if chaosmemory and tonumber(player.chaoticsettings.memory) > 0 then
        for index, value in ipairs(chaosmemory) do
            if neweffecttbl.name == value then
                DebugPrint(player, "did not choose due to being in memory")
                return false
            end
        end
    end

    for key, value in pairs(tbl) do
        if value.name == neweffecttbl.name then
            return false
        end
        if incompatible then
            local found = false
            for key2, value2 in pairs(incompatible) do
                if value.name == value2 then
                    DebugPrint(player, "did not choose due to being incompatible: " .. value.name .. " " .. value2)
                    found = true
                end
            end
            if found then
                return false
            end
        end
    end

    return true
end

-- TODO: SERIALIZE AND DESERIALIZE TABLES

local function CHT_Save1Setting(player, namez, tablez)
    local name = "client/chaoticedition/" .. namez .. ".dat"
    local savefile = io.openlocal(name, "w")
    if tablez then
        savefile:write(Serialize(player.chaoticsettings[namez]))
    else
        savefile:write(tostring(player.chaoticsettings[namez]))
    end
    DebugPrint(player, "Saved to: " .. name .. " with data: " .. tostring(player.chaoticsettings[namez]))
    savefile:close()
end

local function CHT_SaveSettings(player, silent)
    for index, value in ipairs(OptionsTable) do
        if (not value.noset or value.optiontype == "table") then
            if value.playervar then
                CHT_Save1Setting(player, value.name, (value.optiontype == "table"))
            elseif (netgame and not IsPlayerAdmin(player)) or not netgame then
                -- DO SOMETHING HERE
            end
        end
    end
    if not silent then
        addHUDText(player, "\x82Saved settings!")
    end
end

local function CHT_LoadSettings(player, silent)
    for index, value in ipairs(OptionsTable) do
        if not value.noset then
            if value.playervar then
                local name = "client/chaoticedition/" .. value.name .. ".dat"
                local loadfile = io.openlocal(name, "r")
                if loadfile != nil then
                    if value.optiontype == "table" then
                        local tables = Deserialize(loadfile:read("*all"))
                        for key, value in pairs(tables) do
                            print(key)
                        end
                        player.chaoticsettings[value.name] = tables
                        print("TABLE")
                    else
                        player.chaoticsettings[value.name] = loadfile:read("*all")
                        if tonumber(player.chaoticsettings[value.name]) or player.chaoticsettings[value.name] == "0" then
                            player.chaoticsettings[value.name] = tonumber($)
                        elseif player.chaoticsettings[value.name] == "true" or player.chaoticsettings[value.name] == "false" then
                            if player.chaoticsettings[value.name] then
                                player.chaoticsettings[value.name] = true
                            else
                                player.chaoticsettings[value.name] = false
                            end
                        end
                    end
                    DebugPrint(player, "Loaded from: " .. name .. " with data: " .. tostring(player.chaoticsettings[value.name]))
                    loadfile:close()
                end
            elseif (netgame and not IsPlayerAdmin(player)) or not netgame then
                -- DO SOMETHING HERE
            end
        end
    end
    if not silent then
        addHUDText(player, "\x82Loaded settings!")
    end
end


local function GetEffect(player, namez)
    local sindex = 0
    local CopyOf = deepcopy(EffectsTable)
    for index, value in ipairs(CopyOf) do
        if value.name == namez then
            sindex = index
        end
    end
    player.chaotic.StatusTable = CopyOf[sindex]
end

local function ApplyEffect(player, sable)
    if not sable then
        sable = player.chaotic.StatusTable
    end
    S_ChaoticSFX(player, "start", sable)
    sable.duration = $ * TICRATE
    sable.timer = sable.duration
    if sable.addition and amplifymode == "on" then
        sable.timer = $ + (sable.addition * TICRATE)
    end
    sable.started = false
    -- player.chaoticsettings.position
    if sable.type == "Roulette" then
        sable.ytime = 215
        if player.chaoticsettings.position == "top" then
            sable.ytime = -30
        end
    else
        sable.ytime = 175
        if IsMatchMode() then
            sable.ytime = 155
        end
        if player.chaoticsettings.position == "top" then
            sable.ytime = 8
        end
    end

    table.insert(player.chaoticeffects, sable)
    if not sable.nohudtext then
        addHUDText(player, sable.name)
    end
    if sable.type != "Roulette" then
        CONS_Printf(player, "*\x83 applied\x80: " .. sable.name)
        DebugPrint(player, "timer: " .. sable.timer)
    end
    if sable.name != "COMBO TIME!" then -- might not need this anymore?
        player.chaotic.StatusApplied = true
    end

    sable = nil
    player.chaotic.StatusEffect = nil
end

local function TheVibes(player)
    for index, value in ipairs(player.chaoticeffects) do
        -- [[ the vibes v2.0, aka the roulette code ]] --
        if value.type == "Roulette" then
            if not value.started then
                value.name = RandomWord() + index -- ensure it can't conflict
                value.started = true
                value.RouletteTimer = 0
                value.RouletteTimerMax = 12
                value.rerollTimer = 0
                value.StatusRand = P_RandomRange(1, player.chaotic.numberofeffects)
                value.ytime = 215
                value.ymax = 175
                if player.chaoticsettings.position == "top" then
                    value.ytime = -30
                    value.ymax = 8
                end
                if IsMatchMode() then
                    value.ymax = 155
                    value.RouletteTimerMax = 16
                end
            end

            if value.rerollTimer > 0 then
                value.rerollTimer = $ - 1
            end
        
            if value.freeze then
                break
            end

            if player.chaoticsettings.position == "top" then
                value.ytime = $ + 2
                if value.ytime > value.ymax then
                    value.ytime = value.ymax
                end
            else
                value.ytime = $ - 2
                if value.ytime < value.ymax then
                    value.ytime = value.ymax
                end
            end

            if value.rerollTimer <= 0 then
                value.StatusRand = P_RandomRange(1, player.chaotic.numberofeffects)
                if P_RandomChance(vibechance) or value.reroll or player.chaotic.FirstEffect then -- P_RandomChance(vibechance) or
                    player.chaotic.FirstEffect = false
                    if value.reroll then
                        DebugPrint(player, "*\x85 reroll: \x80" .. value.name)
                        value.timer = 0
                        value.reroll = false
                    end
                    local CopyOf = deepcopy(player.chaoticpstable) -- TO CHANGE IF IMPLEMENTING FASTER TABLE
                    value.StatusTable = CopyOf[value.StatusRand]
                    value.StatusEffect = value.StatusTable.name
                    value.sprite = value.StatusTable.sprite
                    value.RouletteTimer = $ + 1
                    -- Lucky! affecting chance
                    if player.chaotic.IsLucky then
                        if value.StatusTable.effect != "bad" then
                            DebugPrint(player, "chance before: " .. value.StatusTable.chance)
                            if value.StatusTable.effect == "good" then
                                value.StatusTable.chance = $/4
                            else
                                value.StatusTable.chance = $/7
                            end
                            DebugPrint(player, "chance after: " .. value.StatusTable.chance)
                        elseif value.StatusTable.duration > 0 then
                            value.StatusTable.duration = $ - 1
                            DebugPrint(player, "timer reduced by 1")
                        end
                    end
                    DebugPrint(player, "*\x82 rolled\x80: " .. value.StatusEffect)
                    if player.chaoticsettings.roulettesfx == "on" then
                        S_StartSoundAtVolume(player.mo, sfx_cdfm46, 80, player)
                    end
            
                    -- [ status chances ] --
                    if value.StatusEffect then -- sfx_s3kd2s
                        if value.RouletteTimer >= value.RouletteTimerMax and player.playerstate != PST_DEAD and P_RandomChance(value.StatusTable.chance)
                        and NonConflictingEffect(player, nil, value.StatusTable) then
                            ApplyEffect(player, value.StatusTable)
                            player.chaotic.StatusTable = nil
                        else
                            value.reroll = true
                            value.rerollTimer = 2
                            return
                        end
                    end
                end
                RemoveEffect(player, index, value, true)
                --player.chaotic.vibecheck = false
            end
        else -- not a roulette, need to adjust ytime value
            if not value.freeze then
                if player.chaoticsettings.position == "top" then
                    value.ytime = 8
                else
                    value.ytime = 175
                    if IsMatchMode() then
                        value.ytime = 155
                    end
                end
            end
        end
    end
end

local function PseudoTable(player)
    if not player.chaoticpstable then
        player.chaoticpstable = {}
    end
    player.chaoticpstable = deepcopy(EffectsTable)

    for index, value in ipairs(player.chaoticpstable) do
        if value.type == "Roulette" then
            --DebugPrint(player, "this is the roulette! free re-roll i guess, you have a 1/65536 to land on this lol")
            table.remove(player.chaoticpstable, index)
        end

        if (value.badmode and amplifymode != "on") then
            --DebugPrint(player, "did not choose due to not being in bad mode")
            table.remove(player.chaoticpstable, index)
        end

        if value.nonet and netgame then
            --DebugPrint(player, "did not choose due to being in a netgame")
            table.remove(player.chaoticpstable, index)
        end

        if value.nomatch and IsMatchMode() then
            --DebugPrint(player, "did not choose due to being in match mode")
            table.remove(player.chaoticpstable, index)
        end

        if value.nobot and player.bot > 0 then
            table.remove(player.chaoticpstable, index)
        end

        if value.name == "Remove Random Effect" and tablelength(player.chaoticeffects) < 2 then
            --DebugPrint(player, "did not choose due to there not being enough effects")
            table.remove(player.chaoticpstable, index)
        end
        if value.name == "Earthquake" and numberinserver() > 2 then
            --DebugPrint(player, "did not choose due to 3+ players")
            table.remove(player.chaoticpstable, index)
        end
        if (value.name == "Anti-Superring" or value.name == "Firework") and player.rings == 0 then
            --DebugPrint(player, "did not choose due no rings")
            table.remove(player.chaoticpstable, index)
        end
        if value.nonights and (mapheaderinfo[gamemap].typeoflevel & TOL_NIGHTS) then
            --DebugPrint(player, "did not choose due to being in NiGHTs")
            table.remove(player.chaoticpstable, index)
        end
        if value.nocart and (player.powers[pw_carry] == CR_MINECART or player.powers[pw_carry] == CR_ROLLOUT or player.powers[pw_carry] == CR_ZOOMTUBE) then
            --DebugPrint(player, "did not choose due to being in cart/rollout/zoomtube")
            table.remove(player.chaoticpstable, index)
        end
        if value.nospecialstage and G_IsSpecialStage(gamemap) then
            --DebugPrint(player, "did not choose due to special stage")
            table.remove(player.chaoticpstable, index)
        end
        if value.noboss and (mapheaderinfo[gamemap].bonustype == 1 or mapheaderinfo[gamemap].bonustype == 2) then
            --DebugPrint(player, "did not choose due to being in a boss stage")
            table.remove(player.chaoticpstable, index)
        end

        if player.chaoticsettings.blacklist then
            for key2, value2 in pairs(player.chaoticsettings.blacklist) do
                if value.name == key2 then
                    --DebugPrint(player, "did not choose due being in blacklist")
                    table.remove(player.chaoticpstable, index)
                end
            end
        end

        if player.chaoticmemory and tonumber(player.chaoticsettings.memory) > 0 then
            for index2, value2 in ipairs(player.chaoticmemory) do
                if value.name == value2 then
                    --DebugPrint(player, "did not choose due to being in memory")
                    table.remove(player.chaoticpstable, index)
                end
            end
        end
    end

    player.chaotic.numberofeffects = tablelength(player.chaoticpstable)
end

addHook("ThinkFrame", function()
    local replace = searchOption("amplify")
    if replace != amplifymode then
        amplifymode = replace
        for player in players.iterate do
            addHUDText(player, "\x81" .. "Amplify Mode\x80 has been set to \x82" .. amplifymode)
        end
    end
end)

addHook("PlayerThink", function(player)
    if not player.chaotic then return end
    -- if netgame then
    if not player.mo then return end
    if player.spectator then return end

    for index, value in ipairs(player.chaotictext) do
        -- starting position, then add 10 for each value
        value.timer = $ + 1

        if value.timer < 2*TICRATE then
            value.flags = V_PERPLAYER|V_ALLOWLOWERCASE
        else
            if (((value.timer-2*TICRATE)*2) % 5 == 0) then
                value.trans = $ + 1
            end
            value.flags = V_PERPLAYER|V_ALLOWLOWERCASE|value.trans<<V_ALPHASHIFT
        end

        if value.timer > value.expire then
            table.remove(player.chaotictext, index)
        end
    end

    TheVibes(player)

    -- roulette object over player's head, incomplete
    --[[for index, value in ipairs(player.chaoticeffects) do
        if not value.object then
            value.object = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z + (player.mo.height + (32*FRACUNIT)), MT_CHAOSROUL)
            value.object.gayer = player.mo
            --value.object.sprite = value.sprite
            value.object.index = index
            value.object.maxindex = tablelength(player.chaoticeffects) + 2
            value.object.offset = 0
            value.object.XCenterVirtual = 0
        elseif value.object.valid then
            value.object.index = index
            value.object.maxindex = tablelength(player.chaoticeffects) + 2
            -- offset = $  + (roulette.index-1)*(20*roulette.scale)
            --value.object.offset = offset
        end
    end]]

    --[[ Faster Roulette Table ]]--

    if (leveltime % 17) == 0 then
        PseudoTable(player)
    end

    if (player.cmd.buttons & BT_CUSTOM2) then
        for index, value in ipairs(player.chaoticpstable) do
            DebugPrint(player, index .. ". " .. value.name)
        end
    end
end)

addHook("PlayerThink", function(player)
    if not player.chaotic then return end
    	if player.chaotic.rerollTimer > 0 then
		player.chaotic.rerollTimer = $ - 1
	end

    if player.spectator then
        return
    end

    -- [[ vibe check ]] --
	if ((not (player.pflags & PF_FINISHED) and not stoppedclock) or player.chaoticsettings.afterfinish == "on")
    and player.playerstate == PST_LIVE and not player.spectator and player.quittime == 0 then
        player.chaotic.timer = $ + 1

        if (not player.chaoticsettings.limit or player.chaoticsettings.limit == 0 or (tablelength(player.chaoticeffects) < player.chaoticsettings.limit)) then
            if not player.chaotic.WaitTime then
                player.chaotic.WaitTime = P_RandomRange(TICRATE*player.chaoticsettings.minwait, TICRATE*player.chaoticsettings.maxwait)
            end
            if player.chaotic.timer > player.chaotic.WaitTime then
                DebugPrint(player, "vibe check")
                player.chaotic.timer = 0
                player.chaotic.WaitTime = nil
                --player.chaotic.vibecheck = true
                GetEffect(player, "Roulette")
                ApplyEffect(player)
            end
        end
    end

    if player.chaoticsettings.effectbtn != "off" and not IsMatchMode() then
        local checkbutton = BtnToName[player.chaoticsettings.effectbtn]
        if (player.cmd.buttons & checkbutton) and not player.chaotic.holdeffectbutton then
            player.chaotic.holdeffectbutton = true

            if player.chaoticsettings.effecttype == "off" then
                player.chaotic.timer = 696969 -- nice
            else
                GetEffect(player, player.chaoticsettings.effecttype)
                if NonConflictingEffect(player, true) then
                    ApplyEffect(player)
                end
            end
        end
        if player.chaotic.holdeffectbutton and (player.cmd.buttons == player.cmd.buttons & ~checkbutton) then
            player.chaotic.holdeffectbutton = false
        end
    end

    if (player.cmd.buttons & BT_CUSTOM1) then
        DebugPrint(player, "CURRENT FX:")
        for index, value in ipairs(player.chaoticeffects) do
            DebugPrint(player, index .. ". " .. value.name .. ": " .. value.timer/TICRATE)
        end
        DebugPrint(player, "TABLE LENGTH: " .. tablelength(player.chaoticeffects))
        DebugPrint(player, "MEMORY:")
        for index, value in ipairs(player.chaoticmemory) do
           DebugPrint(player, index .. ". " .. value)
        end
        DebugPrint(player, "TRANS EFFECTS:")
        for index, value in ipairs(player.chaoticeffectstrans) do
            DebugPrint(player, index .. ". " .. value.name .. ": " .. value.timer/TICRATE)
        end
    end

    --[[ main chaos effect stuffs ]] --
    local Freezer = false
    if (((player.pflags & PF_FINISHED) or stoppedclock) and player.chaoticsettings.afterfinish == "off") or player.playerstate == PST_DEAD then
        Freezer = true
    end

    for index, value in ipairs(player.chaoticeffects) do
        value.timer = $ - 1
        if value.func then
            value.func(player, value, index)
        end

        if Freezer then
            if not value.freeze then
                if value.type != "Roulette" then
                    local fax = deepcopy(value)
                    table.insert(player.chaoticeffectstrans, fax)
                end
                value.freeze = true
            end

            -- code here: if ytime has reached bottom/top, then set timer to 0
            -- also: don't play sounds when timer reaches 0 if value is frozen

            if player.chaoticsettings.position == "top" then
                value.ytime = $ - 2
                if value.ytime < -30 then
                    value.ytime = -30
                    if value.type != "Roulette" then
                        value.timer = 0
                    end
                end
            else
                value.ytime = $ + 2
                if value.ytime > 215 then
                    value.ytime = 215
                    if value.type != "Roulette" then
                        value.timer = 0
                    end
                end
            end
        else
            value.freeze = false
        end
   

        if value.name == "Remove Random Effect" then
            local RANDNUMB = P_RandomRange(1, tablelength(player.chaoticeffects))
            if value.started then
                RemoveEffect(player, index, value)
            else
                RANDNUMB = P_RandomRange(1, tablelength(player.chaoticeffects))
            end

            for index2, value2 in ipairs(player.chaoticeffects) do --inception
                if index2 == RANDNUMB and value2.type != "Roulette" and value2.name != "Remove Random Effect" then
                    value2.timer = 0
                    value.started = true
                    break
                end
            end
        elseif value.name == "Cement Shoes" then
            if not value.started then
                value.started = true
                player.chaotic.CementPrevColor = player.mo.color
                if player.powers[pw_shield] & SH_PROTECTSPIKE then
                    player.chaotic.ShieldAlreadyHasSpike = true
                end
            end
            if not IsMatchMode() then
                player.mo.colorized = true
		        player.mo.color = SKINCOLOR_GREY
            end
            --[[if not player.chaotic.ShieldAlreadyHasSpike and amplifymode != "on" then
			    player.powers[pw_shield] = $ | SH_PROTECTSPIKE
		    end]]
            if (mapheaderinfo[gamemap].typeoflevel & TOL_NIGHTS) then
                player.mo.momx = $ / 3
                player.mo.momy = $ / 3
                player.mo.momz = $ / 4
            else
                if amplifymode == "on" then
                    player.normalspeed = skins[player.mo.skin].normalspeed / 7
                    player.runspeed = skins[player.mo.skin].runspeed / 5
                    player.actionspd = skins[player.mo.skin].actionspd / 9
                    player.thrustfactor = skins[player.mo.skin].thrustfactor / 5
                    player.mindash = FRACUNIT
                    player.maxdash = skins[player.mo.skin].maxdash / 3
                else
                    player.normalspeed = skins[player.mo.skin].normalspeed / 5
                    player.runspeed = skins[player.mo.skin].runspeed / 4
                    player.actionspd = skins[player.mo.skin].actionspd / 7
                    player.thrustfactor = skins[player.mo.skin].thrustfactor / 3
                end

                -- P_SetObjectMomZ(player.mo, -gravity, true)

                if (player.mo.eflags & MFE_VERTICALFLIP) or (player.mo.flags2 & MF2_OBJECTFLIP) then
                    if player.mo.momz > 0 then
                        player.mo.momz = $ + $ / 10
                    else
                        player.mo.momz = $ - $ / 8
                    end
                else
                    if player.mo.momz > 0 then
                        player.mo.momz = $ - $ / 10
                    else
                        player.mo.momz = $ + $ / 8
                    end
                end
            end
            if not P_IsObjectOnGround(player.mo) then
                player.chaotic.CementShoesSfx = true
                if player.mo.momz < -(3*FRACUNIT) and not player.chaotic.CementShoesFallSfx then
                    S_StartSound(player.mo, sfx_s3k51, player)
                    player.chaotic.CementShoesFallSfx = true
                end
            end
            if P_IsObjectOnGround(player.mo) then
                if player.chaotic.CementShoesSfx then
                    S_StartSound(player.mo, sfx_pstop) --previously sfx_doord2
                    if not netgame or (netgame and numberinserver() <= 2) then
                        P_StartQuake(10*FRACUNIT, 5, nil, nil)
                    end
                    player.chaotic.CementShoesSfx = false
                end
                player.chaotic.CementShoesFallSfx = false
            end

            if value.timer <= 0 or player.playerstate == PST_DEAD then -- stop the effect
                RemoveEffect(player, index, value)
                player.normalspeed = skins[player.mo.skin].normalspeed
                player.runspeed = skins[player.mo.skin].runspeed
                player.actionspd = skins[player.mo.skin].actionspd
                player.thrustfactor = skins[player.mo.skin].thrustfactor
                player.mindash = skins[player.mo.skin].mindash
                player.maxdash = skins[player.mo.skin].maxdash
                player.mo.color = player.chaotic.CementPrevColor
                player.mo.colorized = false
                if not player.chaotic.ShieldAlreadyHasSpike then
                    player.powers[pw_shield] = $ & ~SH_PROTECTSPIKE
                end
                P_SpawnShieldOrb(player)
            end


        elseif value.name == "No Abilities" then
            if not value.started then
                value.started = true
                value.OGdrillmeter = player.drillmeter
            end
            
            player.chaotic.CharAbilityNow = true
            player.charability = CA_NONE
            player.charability2 = CA2_NONE
            player.drillmeter = 0
            player.dashmode = 0

            player.pflags = $1 | PF_THOKKED

            if amplifymode == "on" then
                player.charflags = $ | SF_NOJUMPSPIN
                player.charflags = $ | SF_NOJUMPDAMAGE
                value.sprite = "CEICE0"
            end

            if value.timer <= 0 then -- stop the effect
                RemoveEffect(player, index, value)
                player.chaotic.CharAbilityNow = false
                player.charability = skins[player.mo.skin].ability
                player.charability2 = skins[player.mo.skin].ability2
                player.charflags = skins[player.mo.skin].flags
                player.drillmeter = value.OGdrillmeter
                player.chaotic.OGdrillmeter = nil
            end


        elseif value.name == "Giant" then
            if not value.started then
                player.chaotic.OGGiantScale = player.mo.scale
                player.mo.destscale = (FRACUNIT*4)
                value.started = true
            end

            if (player.powers[pw_carry] == CR_MINECART or player.powers[pw_carry] == CR_ROLLOUT) then
                --debugprint("giant but in minecart or rollout rock, aborting")
                value.timer = - 5
            end

            if value.timer <= 0 then
                RemoveEffect(player, index, value)
                player.chaotic.GiantTimer = 0
                player.mo.destscale = FRACUNIT
            end


        elseif value.name == "Tiny" then
            if not value.started then
                player.chaotic.OGTinyScale = player.mo.scale
                if amplifymode == "on" then
                    player.mo.destscale = (FRACUNIT/6)
                else
                    player.mo.destscale = (FRACUNIT/5)
                end
                value.started = true
            end

            if (player.powers[pw_carry] == CR_MINECART or player.powers[pw_carry] == CR_ROLLOUT) then
                --debugprint("tiny but in minecart or rollout rock, aborting")
                value.timer = -5
            end
            if value.timer <= 0 then
                RemoveEffect(player, index, value)
                player.mo.destscale = FRACUNIT
            end


        elseif value.name == "Fling Player" then
            local RANDY = P_RandomChance(FRACUNIT/2)
            local RANDYSLAP
            if amplifymode == "on" then
                if RANDY then
                    P_InstaThrust(player.mo, player.mo.angle, 310*FRACUNIT)
                    if P_IsObjectOnGround(player.mo) then RANDYSLAP = P_RandomChance(FRACUNIT/2 + FRACUNIT/6) end -- 66.6%
                else
                    P_InstaThrust(player.mo, player.mo.angle, -290*FRACUNIT)
                    if P_IsObjectOnGround(player.mo) then RANDYSLAP = P_RandomChance(FRACUNIT/3) end -- 33.3%
                end
            else
                if RANDY then
                    P_InstaThrust(player.mo, player.mo.angle, 210*FRACUNIT)
                    if P_IsObjectOnGround(player.mo) then RANDYSLAP = P_RandomChance(FRACUNIT/2 + FRACUNIT/3) end -- 83.3%
                else
                    P_InstaThrust(player.mo, player.mo.angle, -190*FRACUNIT)
                    if P_IsObjectOnGround(player.mo) then RANDYSLAP = P_RandomChance(FRACUNIT/2 + FRACUNIT/6) end -- 66.6%
                end
            end
            if RANDYSLAP then
                value.name = "Slapped!"
                player.mo.state = S_PLAY_PAIN
                player.climbing = 0
                player.powers[pw_flashing] = TICRATE
                if amplifymode == "on" then
                    player.mo.momz = $ + 35*FRACUNIT
                else
                    player.mo.momz = $ + 20*FRACUNIT
                end
                S_StartSound(player.mo, sfx_kc5b, player)
                S_StartSound(player.mo, sfx_slappy)
            else
                S_StartSound(player.mo, sfx_s3k81, player)
            end

            addHUDText(player, value.name)
            RemoveEffect(player, index, value)


        elseif value.name == "Random Music" then
            S_ChangeMusic(MusicRand(P_RandomRange(1, MusicRand("RANGE"))), true, player)
            RemoveEffect(player, index, value)


        elseif value.name == "Spawn Object" then
            S_StartSound(player.mo, sfx_s3kcas)
            local lengthotable = tablelength(ObjectsTable)
            local RANDY1 = P_RandomRange(1, lengthotable)

            local theobj = ObjectsTable[RANDY1].id
            local name = ObjectsTable[RANDY1].name
            local tic

            local Obj = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z + (player.mo.height + (64*FRACUNIT)), theobj)
            local RANDY2 = P_RandomRange(1, 3)
            Obj.target = player.mo
            --Obj.tracer = player.mo
            Obj.angle = player.mo.angle
            if tic then
                Obj.fuse = tic 
            end
            if Obj.ceilingrover and Obj.ceilingrover.valid then
                if Obj.z > Obj.ceilingrover.bottomheight then
                    Obj.z = Obj.ceilingrover.bottomheight - Obj.height - 4*FRACUNIT
                end
            else
                if Obj.z > Obj.subsector.sector.ceilingheight then
                    Obj.z = Obj.subsector.sector.ceilingheight - Obj.height - 4*FRACUNIT
                end
            end

            if RANDY2 == 1 then
                P_InstaThrust(Obj, Obj.angle, 22*FRACUNIT)
                Obj.angle = $ + ANGLE_180
            elseif RANDY2 == 3 then
                P_InstaThrust(Obj, Obj.angle, -22*FRACUNIT)
            end

            addHUDText(player, "Spawned " .. name)

		-- include super duper rare chance of brak eggman spawning

            RemoveEffect(player, index, value)


        elseif value.name == "I WAS FROZEN TODAY!" then
            if not value.started then
                player.chaotic.FrozenNow = true
                player.chaotic.FrozenPrevColor = player.mo.color
                value.started = true
            end
            player.mo.colorized = true
            player.mo.color = SKINCOLOR_ICY
            if not player.chaotic.FrozenState then
                player.chaotic.FrozenState = player.mo.state
                player.chaotic.FrozenFrame = player.mo.frame
                if player.followmobj then
                    player.chaotic.FollowFrozenState = player.followmobj.state
                    player.chaotic.FollowFrozenFrame = player.followmobj.frame
                end
            end
            if value.timer <= 0 then
                player.mo.color = player.chaotic.FrozenPrevColor
                player.mo.colorized = false
                player.chaotic.FrozenNow = false
                player.chaotic.FrozenState = nil
                player.chaotic.FollowFrozenState = nil
                player.chaotic.FrozenFrame = nil
                player.chaotic.FollowFrozenFrame = nil
                RemoveEffect(player, index, value)
            end


        elseif value.name == "Butter Physics" then
            if not value.started then
                value.started = true
            end

            if amplifymode == "on" then
                player.mo.friction = FRACUNIT + (FRACUNIT/69) -- nice
            else
                player.mo.friction = FRACUNIT + (FRACUNIT/56)
                -- gave me an error about "player.mo" not existing??? why??? this shouldn't be running if the player is dead damnit
            end

            if value.timer <= 0 then
                player.mo.friction = player.chaotic.OGFriction
                RemoveEffect(player, index, value)
            end


        elseif value.name == "Drifting Left" then
            if not value.started then
                value.started = true
                value.adjustvalue = 2
                value.adjustangle = ANG2
                -- CEIDD0
                if P_RandomChance(FRACUNIT/2) then
                    value.sprite = "CEIDD0"
                    value.adjustvalue = -2
                    value.adjustangle = -ANG2
                    addHUDText(player, "Drifting Right")
                else
                    addHUDText(player, "Drifting Left")
                end
            end

            if (mapheaderinfo[gamemap].typeoflevel & TOL_NIGHTS) then
                player.flyangle = $ + value.adjustvalue
                player.anotherflyangle = $ + value.adjustvalue
            else
                if amplifymode == "on" then
                    player.mo.angle = (player.cmd.angleturn << FRACBITS) + value.adjustangle + value.adjustangle
                else
                    player.mo.angle = (player.cmd.angleturn << FRACBITS) + value.adjustangle
                end
            end

            if value.timer <= 0 then
                RemoveEffect(player, index, value)
            end


        elseif value.name == "Negative Rings" then
            if not value.started then
                value.started = true
                player.chaotic.NegativeRingsNow = true
                value.previousrings = player.rings
            end

            if player.rings > value.previousrings then
                local soup = player.rings - value.previousrings
                if amplifymode == "on" then
                    soup = $ * 3
                else
                    soup = $ * 2
                end
                P_GivePlayerRings(player, -soup)
            end

            value.previousrings = player.rings


            if value.timer <= 0 then
                RemoveEffect(player, index, value)
                player.chaotic.NegativeRingsNow = false
            end


        elseif value.name == "2x Rings" then
            if not value.started then
                value.started = true
                player.chaotic.Rings2xNow = true
                value.previousrings = player.rings

                if amplifymode == "off" then
                    addHUDText(player, value.name)
                else
                    addHUDText(player, "4x Rings")
                end
            end
    
            if player.rings > value.previousrings then
                local soup = player.rings - value.previousrings
                if amplifymode == "on" then
                    soup = ($ * 3)
                end
                P_GivePlayerRings(player, soup)
            end
    
            value.previousrings = player.rings

            if value.timer <= 0 then
                RemoveEffect(player, index, value)
                player.chaotic.Rings2xNow = false
            end


        elseif value.name == "Change Character" then
            if not value.started then
                value.started = true
                local RANDYC = P_RandomRange(-5, TICRATE-(TICRATE)/2) -- approximately 17 tics, or half a second
                value.timer = $ - RANDYC
            end

            local skinnum = P_RandomRange(0, numberofskins())
            if (value.timer % 5 == 0) then -- P_RandomRange(4, 6)
                if not R_SkinUsable(player, skinnum) then
                    skinnum = P_RandomRange(0, 2)
                end
                S_StartSoundAtVolume(player.mo, sfx_kc56, 136, player)
                R_SetPlayerSkin(player, skinnum)
            end

            if value.timer <= 0 then
                RemoveEffect(player, index, value)
                if not value.freeze then
                    local dust = P_SpawnMobj(player.mo.x + P_ReturnThrustX(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)),
                    player.mo.y + P_ReturnThrustY(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)), player.mo.z, MT_DUST)
                    dust.scale = FRACUNIT*2
                    dust.state = S_XPLD1
                    addHUDText(player, "Character changed to: " .. skins[player.mo.skin].realname)
                end
            end


        elseif value.name == "Change Color" then
            if not value.started then
                value.started = true

            end

            if (value.timer % 4 == 0) then
                player.mo.color = P_RandomRange(1, #skincolors-1)
                S_StartSoundAtVolume(player.mo, sfx_kc56, 136, player)
                if not IsMatchMode() then
                    player.skincolor = player.mo.color
                end
            end

            if value.timer <= 0 then
                RemoveEffect(player, index, value)
                if not value.freeze then
                    local dust = P_SpawnMobj(player.mo.x + P_ReturnThrustX(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)),
                    player.mo.y + P_ReturnThrustY(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)), player.mo.z, MT_DUST)
                    dust.scale = FRACUNIT*2
                    dust.state = S_XPLD1
                    addHUDText(player, "Color changed to: " .. R_GetNameByColor(player.mo.color))
                end
            end


        elseif value.name == "the ride never ends" then
            if not value.started then
                value.started = true
                local RANDYSPIN = P_RandomChance(FRACUNIT/2)
                if not RANDYSPIN then
                    player.chaotic.SpinOtherWay = true
                end
            end

            if amplifymode == "on" then
                if not player.chaotic.SpinOtherWay then
                    if player.viewrollangle < (ANG30 + ANG30) then
                        player.viewrollangle = $ + (ANG1/4)
                        player.mo.rollangle = $ - (ANG1/4)
                    end
                else
                    if player.viewrollangle > -(ANG30 + ANG30) then
                        player.viewrollangle = $ - (ANG1/4)
                        player.mo.rollangle = $ + (ANG1/4)
                    end
                end
            else
                if not player.chaotic.SpinOtherWay then
                    if player.viewrollangle < (ANG30 + ANG15) then
                        player.viewrollangle = $ + (ANG1/5)
                        player.mo.rollangle = $ - (ANG1/5)
                    end
                else
                    if player.viewrollangle > -(ANG30 + ANG15) then
                        player.viewrollangle = $ - (ANG1/5)
                        player.mo.rollangle = $ + (ANG1/5)
                    end
                end
            end

            if value.timer <= 0 then
                player.chaotic.RestoreViewAngle = true
                RemoveEffect(player, index, value)
            end


        elseif value.name == "Nuke" then
            if player.powers[pw_shield] then
                player.chaotic.NukePrevShield = player.powers[pw_shield]
            end

            P_BlackOw(player)

            if player.chaotic.NukePrevShield then
                player.powers[pw_shield] = player.chaotic.NukePrevShield
                player.chaotic.NukePrevShield = nil
            end

            RemoveEffect(player, index, value)


        elseif value.name == "Darkened Vision" then
            if not value.started then
                value.started = true
            end

            player.chaotic.DarkenedVisionNow = true

            if value.timer <= 0 then
                RemoveEffect(player, index, value)
                player.chaotic.DarkenedVisionNow = false
            end


        elseif value.name == "Invert Controls" then
            if not value.started then
                value.started = true

                local RANDY = P_RandomChance(FRACUNIT/5)

                if RANDY then
                    player.chaotic.InvertCam = true
                end
            end

            player.chaotic.InvertControlsNow = true

            if player.chaotic.InvertCam then -- code taken from "MirrorControls_v2.1.lua" by QuickWolfPT
                if (player.powers[pw_carry] == CR_MINECART or player.powers[pw_carry] == CR_PTERABYTE or player.mo.state == S_PLAY_FIRE) then
                    player.mo.angle = player.drawangle
                    player.chaotic.OldAngle = player.mo.angle
                    return
                end

                if (player.mo.state == S_PLAY_CLING or player.mo.state == S_PLAY_CLIMB or (player.onconveyor == 2 and player.mo.state == S_PLAY_PAIN) or player.powers[pw_carry] == CR_ZOOMTUBE or player.powers[pw_carry] == CR_PTERABYTE) then
                    player.chaotic.OldAngle = player.mo.angle
                    return
                end

                if ((player.pflags & PF_ANALOGMODE) and 1 or 0) then
                    player.chaotic.OldAngle = $ or player.mo.angle

                    local angdiff = (player.cmd.angleturn << FRACBITS) - player.chaotic.OldAngle
                    player.mo.angle = $ - 2 * angdiff

                    player.chaotic.OldAngle = player.mo.angle
                end
            end

            if value.timer <= 0 then
                RemoveEffect(player, index, value)
                player.chaotic.InvertCam = false
                player.chaotic.InvertControlsNow = false
            end


        elseif value.name == "i taketh" then
            S_StartSound(player.mo, sfx_shldls, player)
            player.powers[pw_super] = 0
            P_RemoveShield(player)
            if player.rings > 0 then
                player.powers[pw_invulnerability] = 0
            end
            player.powers[pw_sneakers] = 0
            player.powers[pw_flashing] = 0
            player.powers[pw_tailsfly] = 0
            player.powers[pw_shield] = SH_NONE
            player.powers[pw_nights_superloop] = 0
            player.powers[pw_nights_helper] = 0
            player.powers[pw_nights_linkfreeze] = 0
            if player.rings > 0 then
                if amplifymode == "on" then
                    P_GivePlayerRings(player, -7)
                else
                    P_GivePlayerRings(player, -3)
                end
            end
            player.glidetime = 0
 
            if (mapheaderinfo[gamemap].typeoflevel & TOL_NIGHTS) then
                if player.drillmeter > 0 then
                    player.drillmeter = $ / 2
                elseif player.chaotic.OGdrillmeter > 0
                    player.chaotic.OGdrillmeter = $ / 2
                end
                if player.spheres > 2 then
                    player.spheres = $ - 2
                end
            end
 
            player.climbing = 0
            player.homing = 0
            player.mo.color = player.skincolor
            player.pflags = $1 | PF_THOKKED
            player.pflags = $ & ~PF_GLIDING
            player.pflags = $ & ~PF_SPINNING
            RemoveEffect(player, index, value)
        elseif value.name == "Gravity Coil" then
            if not value.started then
                value.started = true
            end

            if skins[player.mo.skin].jumpfactor >= FRACUNIT then
                player.jumpfactor = (skins[player.mo.skin].jumpfactor + (FRACUNIT/2) + (FRACUNIT/4))
            else
                player.jumpfactor = (FRACUNIT + (FRACUNIT/2) + (FRACUNIT/4))
            end
 
 
            if not P_IsObjectOnGround(player.mo) then
                if (player.mo.eflags & MFE_VERTICALFLIP) or (player.mo.flags2 & MF2_OBJECTFLIP) then
                    if player.mo.momz > 0 then
                        player.mo.momz = $ - $ / 25
                    elseif player.mo.momz < (9*FRACUNIT * player.jumpfactor) then
                        player.mo.momz = $ + $ / 2
                    end
                else
                    if player.mo.momz < 0 then
                        player.mo.momz = $ - $ / 25
                    elseif player.mo.momz < (9*FRACUNIT * player.jumpfactor) then
                        player.mo.momz = $ + $ / 2
                    end
                end
            end

            if value.timer <= 0 then
                RemoveEffect(player, index, value)
                player.jumpfactor = skins[player.mo.skin].jumpfactor
            end


        elseif value.name == "Super Jump" then
            if not value.started then
                value.started = true
            end

            if amplifymode == "on" then
                player.jumpfactor = skins[player.mo.skin].jumpfactor * 5
            else
                player.jumpfactor = skins[player.mo.skin].jumpfactor * 4
            end

            if value.timer <= 0 then
                RemoveEffect(player, index, value)
                player.jumpfactor = skins[player.mo.skin].jumpfactor
            end


        elseif value.name == "the pizza is aggressive" then
            if not value.started then
                value.started = true
            end

            if not player.chaotic.PizzaSpawned then
                GivePizza(player)
            end

            if value.timer <= 0 or player.chaotic.PizzaHitTarget then
                if player.chaotic.PizzaHitTarget then
                    addHUDText(player, "Delivery complete, enjoy your pizza!")
                end
                RemoveEffect(player, index, value)
                --[[if not player.chaotic.PizzaHitTarget then
                    S_StartSound(player.mo, sfx_kc65, player)
                end]]
                player.chaotic.PizzaSpawned = false
                player.chaotic.PizzaRemoval = true
                player.chaotic.PizzaHitTarget = false
            end


        elseif value.name == "Funky Camera" then
            if not value.started then
                value.started = true
                local RANDY = P_RandomChance(FRACUNIT/2)
	
                if RANDY then
                    player.chaotic.CameraDist = true
                    S_StartSound(player.mo, sfx_camout, player)
                else
                    player.chaotic.CameraDist = false
                    S_StartSound(player.mo, sfx_camin, player)
                end
            end

            if amplifymode == "on" then
                if player.chaotic.CameraDist then
                    player.camerascale = skins[player.mo.skin].camerascale * 7
                else
                    player.camerascale = skins[player.mo.skin].camerascale / 9
                end
            else
                if player.chaotic.CameraDist then
                    player.camerascale = skins[player.mo.skin].camerascale * 4
                else
                    player.camerascale = skins[player.mo.skin].camerascale / 5
                end
            end

            if value.timer <= 0 then
                RemoveEffect(player, index, value)
                player.camerascale = skins[player.mo.skin].camerascale
                if player.chaotic.CameraDist then
                    S_StartSound(player.mo, sfx_camin, player)
                else
                    S_StartSound(player.mo, sfx_camout, player)
                end

                player.chaotic.CameraDist = nil
            end


        elseif value.name == "Ghost" then
            if not value.started then
                value.started = true
                player.chaotic.GhostNow = true
            end

            player.pflags = $1 | PF_INVIS
            player.mo.flags = $ | MF_NOCLIPTHING

            if value.timer <= 0 then
                RemoveEffect(player, index, value)
                player.mo.flags = $ & ~MF_NOCLIPTHING
                player.pflags = $ & ~PF_INVIS
                player.powers[pw_flashing] = 1*TICRATE
                player.chaotic.GhostNow = false
                player.chaotic.GhostBegone = true
            end


        elseif value.name == "Change Sprite" then -- VERY BROKEN RIGHT NOW
            if not value.started then
                value.started = true
                player.chaotic.ChangeSpriteNow = true
                local RANDY = P_RandomRange(1, 6)

                if RANDY == 1 then
                    --player.chaotic.SpriteToChangeTo = SPR_POSS
                    player.chaotic.StateToChangeTo = S_SOSS_RUN1
                    player.chaotic.SpriteMOBJ = "tank"
                elseif RANDY == 2 then
                    player.chaotic.StateToChangeTo = S_SSHELL_RUN1
                    player.chaotic.SpriteMOBJ = "spring shell"
                elseif RANDY == 3 then
                    player.chaotic.SpriteToChangeTo = S_COLORRING
                elseif RANDY == 4 then
                    player.chaotic.StateToChangeTo = S_YHORIZ1
                elseif RANDY == 5 then
                    player.chaotic.StateToChangeTo = S_COLORRING
                elseif RANDY == 6 then
                    player.chaotic.SpriteToChangeTo = SPR_EGGM
                end

                local dust = P_SpawnMobj(player.mo.x + P_ReturnThrustX(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)),
                player.mo.y + P_ReturnThrustY(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)), player.mo.z, MT_DUST)
                dust.scale = FRACUNIT*2
                dust.state = S_XPLD1
            end

            if player.chaotic.SpriteMOBJ then
                if player.speed == 0 or player.skidtime > 0 then
                    player.chaotic.SpriteToChangeTo = states[player.chaotic.StateToChangeTo].sprite
                    player.mo.frame = 0
                else
                    player.chaotic.SpriteToChangeTo = nil
                end
            end

            if value.timer <= 0 then
                RemoveEffect(player, index, value)
                player.chaotic.ChangeSpriteNow = false
                player.chaotic.SpriteToChangeTo = nil
                player.chaotic.StateToChangeTo = nil
                player.chaotic.SpriteMOBJ = nil
                player.chaotic.SpriteMOBJ2 = nil
                player.chaotic.Sprung = false
                local dust = P_SpawnMobj(player.mo.x + P_ReturnThrustX(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)), player.mo.y + P_ReturnThrustY(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)), player.mo.z, MT_DUST) 
                dust.scale = FRACUNIT*2
                dust.state = S_XPLD1
            end



        elseif value.name == "Give Good Stuff" then
            S_StartSound(player.mo, sfx_chchng)
	
            if not player.powers[pw_shield] and (P_RandomChance(FRACUNIT/2) or (amplifymode == "on")) then
                player.powers[pw_shield] = SH_PITY
                P_SpawnShieldOrb(player)
                addHUDText(player, "Given a Pity Shield")
                player.chaotic.GivenNiceStuff = true
            end
 
            if P_RandomChance(FRACUNIT/2) then
                player.powers[pw_sneakers] = 20*TICRATE
                --P_RestoreMusic(player)
                addHUDText(player, "Given Super Sneakers")
                player.chaotic.GivenNiceStuff = true
            end
 
            if P_RandomChance(FRACUNIT/3) or (amplifymode == "on" and P_RandomChance(FRACUNIT/2)) then
                if IsMatchMode() then
                    player.powers[pw_invulnerability] = 5*TICRATE
                else
                    player.powers[pw_invulnerability] = 20*TICRATE
                end
                --P_RestoreMusic(player)
                addHUDText(player, "Given Invulernability")
                player.chaotic.GivenNiceStuff = true
            end
 
            if P_RandomChance(FRACUNIT/15) or (amplifymode == "on" and P_RandomChance(FRACUNIT/8)) then
                P_GivePlayerRings(player, 110)
                addHUDText(player, "Given 110 rings")
                player.chaotic.GivenNiceStuff = true
            end
 
            if P_RandomChance(FRACUNIT/99) or (amplifymode == "on" and P_RandomChance(FRACUNIT/70)) then
                P_GivePlayerRings(player, 999)
                if IsMatchMode() then
                    player.powers[pw_invulnerability] = 30*TICRATE
                else
                    player.powers[pw_invulnerability] = 500*TICRATE
                end
                --P_RestoreMusic(player)
                addHUDText(player, "Given 999 rings, and a lot of Invulnerability!")
                player.chaotic.GivenNiceStuff = true
            end
 
            if not player.chaotic.GivenNiceStuff then
                --debugprint("somehow the player got nothing, giving them a bone")
                if IsMatchMode() then
                    player.powers[pw_invulnerability] = 4*TICRATE
                else
                    player.powers[pw_invulnerability] = 15*TICRATE
                end
                P_GivePlayerRings(player, 15)
                addHUDText(player, "Given some invulnerability, and 15 rings")
                --P_RestoreMusic(player)
            end
 
            player.chaotic.GivenNiceStuff = nil
            RemoveEffect(player, index, value)


        elseif value.name == "Earthquake" then
            if not value.started then
                value.started = true
                player.chaotic.QuakeNow = true
                if netgame then chatprint("* " .. player.name .. " has started an earthquake!", true) end
            end

            if amplifymode == "on" then
                P_StartQuake(30*FRACUNIT, 1, nil, nil)
            else
                P_StartQuake(20*FRACUNIT, 1, nil, nil)
            end
	

            if value.timer <= 0 then
                RemoveEffect(player, index, value)
            end


        elseif value.name == "Money Bag" then
            if not value.started then
                value.started = true
                local RANDYC = P_RandomRange(0, TICRATE+(TICRATE)/2) -- approximately 17 tics, or half a second
                value.timer = $ - RANDYC
            end

            local amount = 1
            if amplifymode == "on" then
                amount = P_RandomRange(1, 2)
            end

            if (value.timer % 5 == 0) then
                if G_IsSpecialStage(gamemap) then
                    player.spheres = $ + amount
                    S_StartSound(player.mo, sfx_s3k65)
                else
                    P_GivePlayerRings(player, amount)
                    S_StartSound(player.mo, sfx_itemup)
                end
            end

            if IsMatchMode() then
                local WEPON = P_RandomRange(pw_infinityring, pw_railring)
                player.powers[WEPON] = $ + amount
            end

            if value.timer <= 0 then
                RemoveEffect(player, index, value)
            end


        elseif value.name == "Superring" then
            S_StartSound(player.mo, sfx_itemup)
            if amplifymode == "on" then
                P_GivePlayerRings(player, 20)
            else
                P_GivePlayerRings(player, 10)
            end
            RemoveEffect(player, index, value)


        elseif value.name == "Anti-Superring" then
            S_StartSound(player.mo, sfx_antiri)
            --P_PlayerRingBurst(player, 10)
 
            if player.rings > 0 then
                if amplifymode == "on" then
                    P_GivePlayerRings(player, -20)
                else
                    P_GivePlayerRings(player, -10)
                end
            end
 
            if player.rings == 0 and amplifymode != "on" then
                player.powers[pw_flashing] = 1*TICRATE
            end
            RemoveEffect(player, index, value)


        elseif value.name == "COMBO TIME!" then
            if not value.started then
                value.started = true
                player.chaotic.timer = 696969
            end

            if not value.Doner then
                if player.chaotic.StatusApplied then
                    player.chaotic.timer = 696969
                    player.chaotic.vibecheck = true
                    value.Doner = true
                    S_StartSound(player.mo, sfx_kc4a, player)
                end
            else
                if player.chaotic.StatusApplied then
                    S_StartSound(player.mo, sfx_kc4a, player)
                    RemoveEffect(player, index, value)
                end
            end


        elseif value.name == "Reversed!" then
            player.mo.angle = (player.cmd.angleturn << FRACBITS) + ANGLE_180
            S_StartSound(player.mo, sfx_s3kaa)
            RemoveEffect(player, index, value)


        elseif value.name == "Bouncy Floors" then -- sfx_s3kae
            if not value.started then
                value.started = true
                value.LastMomZ = player.mo.momz
            end

            if player.mo.momz/FRACUNIT == 0 then
                if (player.mo.eflags & MFE_VERTICALFLIP) or (player.mo.flags2 & MF2_OBJECTFLIP) then
                    if value.LastMomZ > 0 then
                        value.BounceTime = true
                    end
                else
                    if value.LastMomZ < 0 then
                        value.BounceTime = true
                    end
                end
            else
                value.LastMomZ = player.mo.momz
            end

            if value.BounceTime then
                if P_IsObjectOnGround(player.mo) then
                    player.pflags = $1 | PF_JUMPSTASIS
                    player.pflags = $1 | PF_THOKKED
                    player.mo.momz = abs(value.LastMomZ)
                    if (player.mo.eflags & MFE_VERTICALFLIP) or (player.mo.flags2 & MF2_OBJECTFLIP) then
                        player.mo.momz = -$
                    end
                    S_StartSoundAtVolume(player.mo, sfx_kc52, 150)
                    value.BounceTime = false
                end
            end

            if value.timer <= 0 then
                RemoveEffect(player, index, value)
                player.pflags = $ & ~PF_JUMPSTASIS
            end


        elseif value.name == "Black Hole" then
            if not value.started then
                value.started = true
            end

            -- thanks to amperbee for providing sample code!
            local m = player.realmo
            local range = 1440*FRACUNIT
            if amplifymode == "on" then
                range = $ + 256 * FRACUNIT
            end
            local range2 = range - (64*FRACUNIT)
            searchBlockmap("objects", function(ref, found)
                if ValidBHObject(found) then
                    if R_PointToDist2(m.x, m.y, found.x, found.y) <= range2 then
                        if found.type == MT_RING or found.type == MT_COIN or found.type == MT_BLUESPHERE then
                            found.tracer = m
                        end
                        -- min = 3 FRACUNIT
                        -- max = 22 FRACUNIT
                        local dist = R_PointToDist2(m.x, m.y, found.x, found.y)
                        if dist < FRACUNIT then
                            dist = FRACUNIT
                        end
                        local PullForce = range/dist * 3
                        if PullForce > 30 then
                            PullForce = 30
                        end
                        PullForce = $ * FRACUNIT

                        BootlegAttract(found, m, PullForce)
                    end
                end
            end, m, m.x-range, m.x+range, m.y-range, m.y+range)

            if value.timer <= 0 then
                RemoveEffect(player, index, value)
            end


        elseif value.name == "Lucky!" then
            if not value.started then
                value.started = true
            end

            player.chaotic.IsLucky = true

            if not IsMatchMode() then
                if value.timer % TICRATE == 0 then -- for every second
                    --((TICRATE/2)+(TICRATE/5))
                    if P_RandomChance(FRACUNIT/20) then
                        local m = player.realmo
                        local range = 512*FRACUNIT
                        if amplifymode == "on" then
                            range = $ * 2
                        end
                        local range2 = range - (64*FRACUNIT)
                        searchBlockmap("objects", function(ref, found)
                            if ValidLuckyObject(found) then
                                if R_PointToDist2(m.x, m.y, found.x, found.y) <= range2 then
                                    found.type = MT_RING_BOX
                                    found.state = mobjinfo[found.type].spawnstate
                                    found.flags = mobjinfo[found.type].flags
                                    addHUDText(player, "Random enemy has changed to ring monitor, lucky!")
                                    return true
                                end
                            end
                        end, m, m.x-range, m.x+range, m.y-range, m.y+range)
                    end
                end
            end

            if value.timer <= 0 then
                RemoveEffect(player, index, value)
                player.chaotic.IsLucky = false
            end


        elseif value.name == "Laggy Controls" then
            if not value.started then
                value.started = true
                player.chaotic.HasLaggyControls = true
            end

            player.chaotic.HasLaggyControls = true

            if value.timer <= 0 then
                RemoveEffect(player, index, value)
                player.chaotic.HasLaggyControls = false
            end


        elseif value.name == "No Turning" then
            if not value.started then
                value.started = true
                value.pangle = player.mo.angle
            end

            player.mo.angle = value.pangle

            if value.timer <= 0 then
                RemoveEffect(player, index, value)
            end


        elseif value.name == "Firework" then
            if not value.started then
                value.started = true
            end

            local additioner = FRACUNIT + (FRACUNIT/2) + (FRACUNIT/4)

            if (mapheaderinfo[gamemap].typeoflevel & TOL_NIGHTS) then
                additioner = $ * 20
            end

            player.mo.momz = $ + additioner
            
            local function SpawnSpark(x, y, z)
                local Sperk = P_SpawnMobj(player.mo.x + x*2, player.mo.y + y*3, player.mo.z + z/4, MT_BOXSPARKLE)
                Sperk.scale = player.mo.scale
                Sperk.momx = x * 2
                Sperk.momy = y * 2
                Sperk.momz = z * 10
            end

            for i=6,1,-1 do
                SpawnSpark(P_RandomFixed(), P_RandomFixed(), P_RandomFixed())
            end

            if value.timer <= 0 then
                RemoveEffect(player, index, value)
                --S_StartSound(player.mo, sfx_kc65, player)
                
                for i=69,1,-1 do -- nice
                    SpawnSpark(P_RandomFixed()*4, P_RandomFixed()*5, P_RandomFixed()*6)
                end
                local amnt = FRACUNIT/2-FRACUNIT/5
                if amplifymode == "on" then
                    amnt = FRACUNIT/2+FRACUNIT/3
                end
                if IsMatchMode() then
                    amnt = $ / 2
                end
                if P_RandomChance(amnt) then
                    local explode = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_TNTBARREL)
                    explode.state = S_TNTBARREL_EXPL1
                end
            end


        elseif value.name == "Moonjump" then
            if not value.started then
                value.started = true
            end

            player.charability = CA_DOUBLEJUMP
            player.pflags = $ & ~PF_THOKKED

            if value.timer <= 0 then
                RemoveEffect(player, index, value)
                player.charability = skins[player.mo.skin].ability
            end


        elseif value.name == "Zero Gravity" then
            if not value.started then
                value.started = true
                value.laststoredmom = player.mo.momz
            end

            if (player.mo.eflags & MFE_VERTICALFLIP) or (player.mo.flags2 & MF2_OBJECTFLIP) then
                if value.laststoredmom < player.mo.momz then
                    player.mo.momz = value.laststoredmom
                end
            else
                if value.laststoredmom > player.mo.momz then
                    player.mo.momz = value.laststoredmom
                end
            end

            value.laststoredmom = player.mo.momz


            if value.timer <= 0 then
                RemoveEffect(player, index, value)

            end


        elseif value.name == "Busted Controller" then
            if not value.started then
                value.started = true
                value.holdtime = 0
            end

            player.chaotic.BustedController = true

            if value.timer <= 0 then
                RemoveEffect(player, index, value)
                player.chaotic.BustedController = false
            end


        elseif value.name == "Windy Day" then
            if not value.started then
                value.angle = player.mo.angle-ANGLE_180
                value.changeang = 0
                value.thrust = FRACUNIT
                value.posneg = false
                value.started = true
            end

            if value.changeang > 0 then
                if value.posneg then
                    value.angle = $ - (ANG2 + ANG1)
                else
                    value.angle = $ + ANG2 + ANG1
                end
                value.changeang = $ - 1
                value.thrust = P_RandomRange(FRACUNIT/4, FRACUNIT) + P_RandomFixed()/3
                if amplifymode then
                    value.thrust = $ + P_RandomFixed()/5
                end
            else
                value.changeang = 0
                value.posneg = not value.posneg
                if P_RandomChance(FRACUNIT/5) then
                    value.changeang = P_RandomRange(5, 25)
                    if amplifymode then
                        value.changeang = $ + 2
                    end
                    --DebugPrint(player, "Changing Angle!")
                end
            end


            P_Thrust(player.mo, value.angle, value.thrust)


            if value.timer <= 0 then
                RemoveEffect(player, index, value)

            end

            
        elseif value.name == "Incline Problems" then
            if not value.started then
                value.started = true
                value.ogfriction = player.mo.friction
            end

            if P_IsObjectOnGround(player.mo) and player.mo.standingslope then
                player.mo.friction = 23*FRACUNIT/32
            else
                player.mo.friction = value.ogfriction
            end

            if value.timer <= 0 then
                RemoveEffect(player, index, value)
                player.mo.friction = value.ogfriction
            end


        elseif value.name == "Bleeding" then
            local function SetTime()
                if amplifymode == "on" then
                    value.timeseed = P_RandomRange(TICRATE/4, TICRATE-TICRATE/5)
                else
                    value.timeseed = P_RandomRange(TICRATE/2, TICRATE+(TICRATE/5))
                end
            end
            if not value.started then
                value.started = true
                SetTime()
            end

            -- every second, player is drained one ring, goes invicible for a split second, and a ring pops out of them

            local function Bleed(silent)
                if player.rings > 0 then
                    if player.powers[pw_flashing] == 0 then
                        player.powers[pw_flashing] = TICRATE/3
                    end
                    local z = (player.mo.z + player.mo.height) + 12*FRACUNIT
                    if (player.mo.eflags & MFE_VERTICALFLIP) then
                        z = (player.mo.z - player.mo.height) - 12*FRACUNIT
                    end
                    local ring = P_SpawnMobj(player.mo.x, player.mo.y, z, MT_FLINGRING)
                    ring.fuse = 5*TICRATE
                    ring.target = player.mo
                    ring.destscale = player.mo.destscale
                    ring.scale = player.mo.scale
                    ring.angle = player.mo.angle

                    P_GivePlayerRings(player, -1)
                    if not silent then
                        S_StartSound(player.mo, sfx_antiri)
                        if P_RandomChance(FRACUNIT/4) then
                            SetTime()
                        end
                    end

                    if (player.mo.eflags & MFE_VERTICALFLIP) then
                        ring.momz = player.mo.momz - player.mo.scale*8
                    else
                        ring.momz = player.mo.momz + player.mo.scale*8
                    end

                    ring.momx = player.mo.momx - (P_RandomRange(-6, 6)*player.mo.scale)
                    ring.momy = player.mo.momy - (P_RandomRange(-6, 6)*player.mo.scale)
                end
            end

            if value.timer % value.timeseed == 0 then
                Bleed()
                if amplifymode == "on" then
                    Bleed(true)
                end
            end

            if value.timer <= 0 then
                RemoveEffect(player, index, value)

            end


        elseif value.name == "PLACEHOLDER" then
            if not value.started then
                value.started = true
            end


            if value.timer <= 0 then
                RemoveEffect(player, index, value)

            end



        end
    end
    -- [[ Misc. Stuff ]] --
    if player.chaotic.RestoreViewAngle then
        if player.chaotic.SpinOtherWay then
            if player.viewrollangle < 0
                player.viewrollangle = $ + (ANG10/4)
                player.mo.rollangle = $ - (ANG10/4)
            else
                player.viewrollangle = 0
                player.mo.rollangle = 0
                player.chaotic.RestoreViewAngle = false
                player.chaotic.SpinOtherWay = false
            end
        else
            if player.viewrollangle > 0 then
                player.viewrollangle = $ - (ANG10/4)
                player.mo.rollangle = $ + (ANG10/4)
            else
                player.viewrollangle = 0
                player.mo.rollangle = 0
                player.chaotic.RestoreViewAngle = false
            end
        end
    end

	if player.chaotic.ReadHelp then
		player.chaotic.ReadHelpTimer = $ + 1
		if player.chaotic.ReadHelpTimer > 8*TICRATE
			player.chaotic.ReadHelp = false
			player.chaotic.ReadHelpTimer = 0
		end
	end

    -- KEEP AT BOTTOM!
    if player.chaotic.StatusApplied then
        player.chaotic.StatusApplied = false
    end

    if Freezer then
        for index, value in ipairs(player.chaoticeffects) do
            value.timer = 0
        end
    end
end)


addHook("PostThinkFrame", function()
    for player in players.iterate do
        if not player.chaotic then return end
		if player.chaotic.GhostNow then  -- thanks to Faye on the SRB2 discord for discovering how to get MF2_SHADOW working
			player.mo.flags2 = $ | MF2_SHADOW
		end
		if player.chaotic.GhostBegone then
			player.mo.flags2 = $ & ~MF2_SHADOW
			player.chaotic.GhostBegone = false
		end

		if player.chaotic.FrozenNow and not player.chaotic.ChangeSpriteNow and player.chaotic.FrozenState then
			player.mo.state = player.chaotic.FrozenState
			player.mo.frame = player.chaotic.FrozenFrame
			player.powers[pw_nocontrol] = 1
			if player.followmobj 
            and player.chaotic.FollowFrozenState and player.chaotic.FollowFrozenFrame then -- redundancy beacuse script errord during testing for some unknown reason
				player.followmobj.state = player.chaotic.FollowFrozenState
				player.followmobj.frame = player.chaotic.FollowFrozenFrame
			end
		end


        if player.chaotic.ChangeSpriteNow then -- SPR_SSWY, S_YHORIZ1
			if player.chaotic.SpriteToChangeTo then
				player.mo.sprite = player.chaotic.SpriteToChangeTo
				player.mo.frame = 0
			else
				player.mo.sprite = states[player.chaotic.StateToChangeTo].sprite
				player.mo.frame = states[player.chaotic.StateToChangeTo].frame
				player.chaotic.StateToChangeToTimer = $ + 1
				if player.chaotic.StateToChangeTo == S_COLORRING then
					if player.chaotic.StateToChangeToTimer <= states[S_RING].var1 then
						player.mo.frame = player.chaotic.StateToChangeToTimer
					else
						player.chaotic.StateToChangeToTimer = 0
					end
				end
	
	
				if player.chaotic.StateToChangeTo == S_SOSS_RUN1 then
					if player.chaotic.StateToChangeToTimer >= states[S_SOSS_RUN1].tics then
						player.chaotic.StateToChangeTo = S_SOSS_RUN2
						player.chaotic.StateToChangeToTimer = 0
					end
				elseif player.chaotic.StateToChangeTo == S_SOSS_RUN2 then
					if player.chaotic.StateToChangeToTimer >= states[S_SOSS_RUN2].tics then
						player.chaotic.StateToChangeTo = S_SOSS_RUN3
						player.chaotic.StateToChangeToTimer = 0
					end
				elseif player.chaotic.StateToChangeTo == S_SOSS_RUN3 then
					if player.chaotic.StateToChangeToTimer >= states[S_SOSS_RUN3].tics then
						player.chaotic.StateToChangeTo = S_SOSS_RUN4
						player.chaotic.StateToChangeToTimer = 0
					end
				elseif player.chaotic.StateToChangeTo == S_SOSS_RUN4 then
					if player.chaotic.StateToChangeToTimer >= states[S_SOSS_RUN4].tics then
						player.chaotic.StateToChangeTo = S_SOSS_RUN5
						player.chaotic.StateToChangeToTimer = 0
					end
				elseif player.chaotic.StateToChangeTo == S_SOSS_RUN5 then
					if player.chaotic.StateToChangeToTimer >= states[S_SOSS_RUN5].tics then
						player.chaotic.StateToChangeTo = S_SOSS_RUN6
						player.chaotic.StateToChangeToTimer = 0
					end
				elseif player.chaotic.StateToChangeTo == S_SOSS_RUN6 then
					if player.chaotic.StateToChangeToTimer >= states[S_SOSS_RUN6].tics then
						player.chaotic.StateToChangeTo = S_SOSS_RUN1
						player.chaotic.StateToChangeToTimer = 0
					end
				end
	
				if player.chaotic.StateToChangeTo == S_SSHELL_RUN1 then
					if player.chaotic.StateToChangeToTimer >= states[S_SSHELL_RUN1].tics then
						player.chaotic.StateToChangeTo = S_SSHELL_RUN2
						player.chaotic.StateToChangeToTimer = 0
					end
				elseif player.chaotic.StateToChangeTo == S_SSHELL_RUN2 then
					if player.chaotic.StateToChangeToTimer >= states[S_SSHELL_RUN2].tics then
						player.chaotic.StateToChangeTo = S_SSHELL_RUN3
						player.chaotic.StateToChangeToTimer = 0
					end
				elseif player.chaotic.StateToChangeTo == S_SSHELL_RUN3 then
					if player.chaotic.StateToChangeToTimer >= states[S_SSHELL_RUN3].tics then
						player.chaotic.StateToChangeTo = S_SSHELL_RUN4
						player.chaotic.StateToChangeToTimer = 0
					end
				elseif player.chaotic.StateToChangeTo == S_SSHELL_RUN4 then
					if player.chaotic.StateToChangeToTimer >= states[S_SSHELL_RUN4].tics then
						player.chaotic.StateToChangeTo = S_SSHELL_RUN1
						player.chaotic.StateToChangeToTimer = 0
					end
				end
	
				if player.chaotic.StateToChangeTo == S_YHORIZ1 and player.chaotic.Sprung then
					if player.chaotic.StateToChangeToTimer >= states[S_YHORIZ1].tics then
						player.chaotic.StateToChangeTo = S_YHORIZ2
						player.chaotic.StateToChangeToTimer = 0
					end
				elseif player.chaotic.StateToChangeTo == S_YHORIZ2 then
					if player.chaotic.StateToChangeToTimer >= states[S_YHORIZ2].tics then
						player.chaotic.StateToChangeTo = S_YHORIZ3
						player.chaotic.StateToChangeToTimer = 0
					end
				elseif player.chaotic.StateToChangeTo == S_YHORIZ3 then
					if player.chaotic.StateToChangeToTimer >= states[S_YHORIZ3].tics then
						player.chaotic.StateToChangeTo = S_YHORIZ4
						player.chaotic.StateToChangeToTimer = 0
					end
				elseif player.chaotic.StateToChangeTo == S_YHORIZ4 then
					if player.chaotic.StateToChangeToTimer >= states[S_YHORIZ4].tics then
						player.chaotic.StateToChangeTo = S_YHORIZ5
						player.chaotic.StateToChangeToTimer = 0
					end
				elseif player.chaotic.StateToChangeTo == S_YHORIZ5 then
					if player.chaotic.StateToChangeToTimer >= states[S_YHORIZ5].tics then
						player.chaotic.StateToChangeTo = S_YHORIZ6
						player.chaotic.StateToChangeToTimer = 0
					end
				elseif player.chaotic.StateToChangeTo == S_YHORIZ6 then
					if player.chaotic.StateToChangeToTimer >= states[S_YHORIZ6].tics then
						player.chaotic.StateToChangeTo = S_YHORIZ7
						player.chaotic.StateToChangeToTimer = 0
					end
				elseif player.chaotic.StateToChangeTo == S_YHORIZ7 then
					if player.chaotic.StateToChangeToTimer >= states[S_YHORIZ7].tics then
						player.chaotic.StateToChangeTo = S_YHORIZ8
						player.chaotic.StateToChangeToTimer = 0
					end
				elseif player.chaotic.StateToChangeTo == S_YHORIZ8 then
					if player.chaotic.StateToChangeToTimer >= states[S_YHORIZ8].tics then
						player.chaotic.StateToChangeTo = S_YHORIZ1
						player.chaotic.StateToChangeToTimer = 0
						player.chaotic.Sprung = false
					end
				end
			end
		end
    end
end)

addHook("PreThinkFrame", function()

    for player in players.iterate do
        if not player.chaotic then return end
        if player.chaotic.InvertControlsNow then
            player.cmd.forwardmove = -($)
            player.cmd.sidemove = -($)
        end

        if IsMatchMode() and player.chaotic.CharAbilityNow then
            if player.cmd.buttons & BT_ATTACK then
                player.cmd.buttons = $ & ~BT_ATTACK
                player.cmd.buttons = $ | BT_FIRENORMAL
            end
        end

        if player.chaotic.FrozenNow then
            player.cmd.forwardmove = 0
            player.cmd.sidemove = 0
            player.cmd.buttons = $ & ~BT_JUMP
            player.cmd.buttons = $ & ~BT_SPIN
        end

        if player.chaotic.HasLaggyControls then
            if not player.LagControlT then
                player.LagControlT = {}
            end
            local function addLag(tbl)
                table.insert(player.LagControlT, tbl)
            end


            -- TODO: Detect letting go of a button, if so, do not add to the table or something. Need to figure this all out.
            if player.cmd.buttons & BT_JUMP then
                addLag{button = "Jump", cmd = BT_JUMP, tics = 1}
            elseif player.cmd.buttons & BT_SPIN then

            end
        end

        if player.chaotic.BustedController then
            if player.chaotic.holdbutton then
                player.cmd.buttons = $ | player.chaotic.holdbutton
                player.chaotic.holdtime = $ - 1
                if player.chaotic.holdtime <= 0 then
                    player.chaotic.holdbutton = nil
                end
            else
                if P_RandomChance(FRACUNIT/5) then
                    if P_RandomChance(FRACUNIT/2) then -- or not P_IsObjectOnGround(player.mo) 
                        player.chaotic.holdbutton = BT_JUMP
                        DebugPrint(player, "press jump")
                    else
                        player.chaotic.holdbutton = BT_SPIN
                        DebugPrint(player, "press spin")
                    end
                    player.chaotic.holdtime = P_RandomRange(2, TICRATE - TICRATE/4)
                    DebugPrint(player, "duration: " .. player.chaotic.holdtime)
                end
            end
            
            if P_RandomChance(FRACUNIT/3) then
                if P_RandomChance(FRACUNIT/2) then
                    player.cmd.forwardmove = -50
                else
                    player.cmd.forwardmove = 50
                end
            end

            if P_RandomChance(FRACUNIT/15) then
                player.mo.angle = (player.cmd.angleturn << FRACBITS) + (P_RandomRange(-60, 60)*ANG1)
            end
        end
    end
end)

addHook("MobjDamage", function(mo)
    local player = mo.player
    if player.chaotic.IsLucky and P_RandomChance(FRACUNIT/5) then
        player.powers[pw_flashing] = 2*TICRATE
        S_StartSound(player.mo, sfx_s3k7a)
        addHUDText(player, "Dodged a hit, lucky!")
        return true
    end
end, MT_PLAYER)

addHook("PlayerSpawn", function(player)
	if not player.mo then
		return
	end

	if not player.chaotic then
		player.chaotic = {}
        player.chaoticeffects = {}
        player.chaoticeffectstrans = {} -- trans rights!
        player.chaoticmemory = {}
        player.chaotictext = {}
        player.chaotic.ReadHelp = true
        player.chaotic.ReadHelpTimer = 0
	end

    player.chaotic.BustedController = false
    player.chaotic.CharAbilityNow = false

    if not player.chaoticsettings then
        player.chaoticsettings = {}
        for index, value in ipairs(OptionsTable) do
            if value.playervar then
                --local CopyOf = deepcopy(OptionsTable)
                --table.insert(player.chaoticsettings, CopyOf[index])
                player.chaoticsettings[value.name] = value.set
            end
        end
        local name = "client/chaoticedition/autoload.dat"
        local loadfile = io.openlocal(name, "r")
        if loadfile != "off" then
            CHT_LoadSettings(player, true)
        end
        if player.chaoticsettings.ReadHelp then
            player.chaotic.ReadHelp = false
        end
    end

	if not player.chaotic.OGFriction then
		player.chaotic.OGFriction = player.mo.friction
	end

    if player.chaotic.GhostNow then
        player.chaotic.GhostNow = false
        player.chaotic.GhostBegone = true
        player.pflags = $ & ~PF_INVIS
    end

    if player.bot == 0 then
        player.chaotic.timer = 0
    elseif not player.chaotic.timer then
        player.chaotic.timer = 0
    end
    player.chaotic.rerollTimer = 0
    player.chaotic.reroll = false
    --player.chaotic.RouletteTimer = 0

    player.mo.scale = FRACUNIT

    player.chaotic.InvertCam = nil

    player.chaotic.GivenNiceStuff = nil
    player.chaotic.StateToChangeToTimer = 0
    player.chaotic.ChangeSpriteNow = false
    player.chaotic.SpriteToChangeTo = nil
    player.chaotic.StateToChangeTo = nil
    player.chaotic.SpriteMOBJ = nil
    player.chaotic.SpriteMOBJ2 = nil
    player.chaotic.Sprung = false

    player.chaotic.StatusEffect = nil

    player.chaotic.NegativeRingsNow = false
    player.chaotic.FrozenNow = false
    player.chaotic.DarkenedVisionNow = false
    player.chaotic.InvertControlsNow = false
    player.chaotic.GhostNow = false

    player.chaotic.SpinOtherWay = nil
    player.chaotic.RestoreViewAngle = nil
    player.chaotic.PizzaHitTarget = nil
    player.chaotic.PizzaRemoval = true
    player.chaotic.CameraDist = nil

    local translength = tablelength(player.chaoticeffectstrans) --owo

    if translength > 0 and player.chaoticsettings.carryover == "on" then
        for index, value in ipairs(player.chaoticeffectstrans) do
            local fax = deepcopy(value)
            if not fax.nospawn then
                table.insert(player.chaoticeffects, fax)
            end
        end
        player.chaoticeffectstrans = {}
    end

    for index, value in ipairs(player.chaoticeffects) do
        if value.object then
            if value.object.valid then
                P_RemoveMobj(value.object)
            else
                value.object = nil
            end
        end
        if value.nospawn then
            RemoveEffect(player, index, value)
        end
    end

    PseudoTable(player)
end)

-- [[ anti-ring and x2 ring ]] --

local function NegativeRings(ring, plyr) local player = plyr.player
	if player.chaotic.NegativeRingsNow then
		--[[if amplifymode == "on" then
			P_GivePlayerRings(player, -2)
		else
			P_GivePlayerRings(player, -1)
		end]]
		P_KillMobj(ring)
		if ring.type == MT_RING then
			S_StartSound(ring, sfx_antiri)
		else
			S_StartSound(ring, sfx_antico)
		end
		ring.shadowscale = 0
        P_GivePlayerRings(player, 1)
		return true
	--[[elseif player.chaotic.Rings2xNow or (player.chaotic.IsLucky and P_RandomChance(FRACUNIT/6)) then
		if amplifymode == "on" then
			P_GivePlayerRings(player, 3)
		else
			P_GivePlayerRings(player, 1)
		end
		return false
	elseif player.chaotic.RandomSFXNow -- sad thing is that this is netplay safe, but the rest of the code for random sfx isn't :(
		P_GivePlayerRings(player, 1)
		P_KillMobj(ring)
		S_StopSound(ring)
		S_StartSound(ring, SoundRand())
		ring.shadowscale = 0
		return true]]
	else
		return false
	end
end

addHook("TouchSpecial", NegativeRings, MT_RING)
addHook("TouchSpecial", NegativeRings, MT_COIN)

local function darkvisionhud(v, player)
    if not player.chaotic then return end
	if v.TransBack == nil
		v.TransBack = 10
	end
	
	if player.chaotic.DarkenedVisionNow then
		if (leveltime % 2 == 0) and ((amplifymode == "on" and v.TransBack > 2) or (amplifymode == "off" and v.TransBack > 3)) then
			v.TransBack = $ - 1
		end
	else
		if (leveltime % 2 == 0) and v.TransBack < 10 then
			v.TransBack = $ + 1
		end
	end
	v.TransFlags = v.TransBack<<V_ALPHASHIFT
	v.drawScaled(0,200 , FRACUNIT*800, v.cachePatch("CEIBD0"), v.TransFlags|V_SNAPTOLEFT|V_SNAPTOTOP)
end

hud.add(darkvisionhud, "game")

local rainbowcolors = { -- Rainbow Table, feel free to use yourself if you credit birbhorse/EeveeEuphoria!
"35",
"34",
"56",
"55",
"54",
"53",
"52",
"51",
"50",
"74",
"73",
"188",
"96",
"112",
"122",
"255",
"123",
"124",
"134",
"135",
"136",
"137",
"150",
"151",
"152",
"167",
"166",
"165",
"184",
"183",
"182",
"181",
"202",
"203",
"204",
"210",
"211",
"212"
}


local function timerhud(v,player)
    if player.chaoticsettings.timerbar == "off" then
        return
    end
    if (((player.pflags & PF_FINISHED) or stoppedclock) and player.chaoticsettings.afterfinish == "off") or player.playerstate == PST_DEAD then
        return
    end
    local width = 320
    local height = 3
    local x = 0
    local y = 200-height
    local SNAP = V_SNAPTOBOTTOM

    local MaxTime
    if not player.chaotic.WaitTime then
        MaxTime = 10*TICRATE
    else
        MaxTime = player.chaotic.WaitTime
    end


    if player.chaoticsettings.position == "top" then
        y = 0
        SNAP = V_SNAPTOTOP
    end

    if not v.color then
        v.color = 0
    end

    local screenwidth = 320
    local XCenter = ((screenwidth/2) - (width/2))

    -- player.chaotic.timer > player.chaotic.WaitTime

    local soup = FixedDiv(player.chaotic.timer, MaxTime)
    local souper = FixedMul(width, soup)

    if player.chaoticsettings.rainbow == "on" then
        local tlength = tablelength(rainbowcolors)
        v.color = $ + 1
        if v.color > tlength
            v.color = 1
        end
        v.colorset = tonumber(rainbowcolors[v.color])
    else
        v.colorset = 36
    end

    v.drawFill(XCenter, y, width, height, 31|V_PERPLAYER|SNAP)
    v.drawFill(XCenter, y, souper, height, v.colorset+5|V_PERPLAYER|SNAP) -- filling timer bar, before the two pixel gap
    v.drawFill(XCenter, y, souper-2, height, v.colorset|V_PERPLAYER|SNAP) -- the actual color for the timer bar

end

hud.add(timerhud, "game")

addHook("MobjThinker", function(aggressive) -- this code is directly lifted from my MMT script, i have no shame
	if aggressive.valid and aggressive.target and aggressive.target.valid then
		if not aggressive.initializedtime then
			aggressive.shadowscale = 3*FRACUNIT
			if aggressive.sizeoption == "large" then
				aggressive.speed = 44*FRACUNIT
				aggressive.scale = FRACUNIT + FRACUNIT*(2/4)
			elseif aggressive.sizeoption == "medium" then
				aggressive.speed = 38*FRACUNIT
				aggressive.scale = FRACUNIT
			elseif (aggressive.sizeoption == "small" or aggressive.sizeoption == "personal") then
				aggressive.speed = (28*3)*FRACUNIT
				aggressive.scale = FRACUNIT/4
			elseif (aggressive.sizeoption == "xl" or aggressive.sizeoption == "extra large" or aggressive.sizeoption == "extralarge") then
				aggressive.speed = 49*FRACUNIT
				aggressive.scale = FRACUNIT*2
			else
				aggressive.speed = (36*2)*FRACUNIT
				aggressive.scale = FRACUNIT/2
			end
			aggressive.initializedtime = true
		end
		if not aggressive.rollanglesprite then
		    aggressive.rollanglesprite = 0
		end
		aggressive.rollanglesprite = $ + ANG15
		aggressive.rollangle = aggressive.rollanglesprite

		if (R_PointToDist2(aggressive.target.x, aggressive.target.y, aggressive.x, aggressive.y))/FRACUNIT < 35 and
        (abs((aggressive.target.z - aggressive.z)/FRACUNIT) < 25) and not (aggressive.target.player and aggressive.target.player.playerstate == PST_DEAD)
        and not aggressive.Remove then
			if amplifymode == "on" then
				P_RemoveShield(aggressive.target.player)
				aggressive.target.player.powers[pw_invulnerability] = 0
				P_DamageMobj(aggressive.target, aggressive, aggressive, DMG_SPIKE)
			else
				P_DamageMobj(aggressive.target, aggressive, aggressive, DMG_SPIKE)
			end

			S_StartSound(aggressive.target, sfx_chapki)
			if aggressive.state == S_CHAOSEXPLOSPIZZA then
				local Explode = P_SpawnMobj(aggressive.target.x, aggressive.target.y, aggressive.target.z, MT_BOSSEXPLODE)
				S_StartSound(explode, sfx_bkpoof)
				Explode.destscale = 3*FRACUNIT
				Explode.scalespeed = FRACUNIT/7
			end
			aggressive.target.player.chaotic.PizzaHitTarget = true
			aggressive.Remove = true
		end

		if aggressive.target and not aggressive.target.player.chaotic.PizzaRemoval then
			A_HomingChase(aggressive, aggressive.speed, 0)
		else
			P_RemoveMobj(aggressive)
		end
	else
		P_RemoveMobj(aggressive)
	end
end, MT_JIMBO)


--[[ TODO:
    - MAKE SURE ALL CALCULATIONS INVOLVING FRACUNIT USE "roulette.scale" INSTEAD!
    - Get icons to work, somehow.
    - Center everything
    - Make it so it's invisible to the player, but visible to everyone else.
        - [?] In the object's thinker, check if [ consoleplayer ] is the player who shouldn't see it. If true, add [ MF2_DONTDRAW ], if false, remove
    - [ ] centering logic: do it like normal (i.e. for each index, add a certain amount), but take the total amount of indexes and do math to offset it


]]

addHook("MobjThinker", function(roulette)
    if roulette.valid and roulette.gayer.valid then
        local player = roulette.gayer.player
        -- calculate angle, then change x and y acccordingly to the index
        roulette.angle = player.mo.angle
        roulette.scale = player.mo.scale
        P_TeleportMove(roulette, player.mo.x, player.mo.y, player.mo.z + (player.mo.height + (32*roulette.scale)))



        local predictedlength = (roulette.maxindex*17) + ((roulette.maxindex-1)*3)
        local screenwidth = player.mo.radius/roulette.scale * 2
    
        local XCenter = ((screenwidth/2) - (predictedlength/2))
    
        --[[if roulette.XCenterVirtual != XCenter then
            if roulette.XCenterVirtual < XCenter then
                roulette.XCenterVirtual = $ + 1
            elseif v.XCenterVirtual > XCenter then
                roulette.XCenterVirtual = $ - 1
            end
        end]]

        roulette.XCenterVirtual = XCenter

        P_InstaThrust(roulette, roulette.angle + ANGLE_90, (roulette.XCenterVirtual*roulette.scale)+((20*roulette.index)*roulette.scale)) -- do offest


        -- !!! HIDING CODE, UNCOMMENT AND TEST IN NETPLAY!!!!


        --[[if displayplayer == consoleplayer then
            roulette.flags2 = $|MF2_DONTDRAW
        else
            roulette.flags2 = $ & ~MF2_DONTDRAW
        end]]
    end
end, MT_CHAOSROUL)

local function PrintHelpForCategory(player, cat)
    CONS_Printf(player, "\x82--------\x83 " .. cat .. "\x82 --------")
    for index, value in ipairs(OptionsTable) do
        if not value.hide and value.cat == cat then
            local setvalue = ""
            local options = ""
            if not value.noset and not value.hideoptions and not value.seperateoptions then
                for i2, v2 in ipairs(value.options) do
                    local slash = "\x80/"
                    if value.optiontype == "number" then
                        slash = "\x80-"
                    end
                    if i2 == tablelength(value.options) then
                        slash = ""
                    end
                    options = $ .. "\x87" .. v2 .. slash
                end
            end
            if not value.noset and not value.hideoptions then
                setvalue = "\x81 set to: \x83" .. value.set
                if value.playervar and player.chaoticsettings then
                    setvalue = "\x81 set to: \x83" .. player.chaoticsettings[value.name]
                end
            end
            local parenth1 = "\x87 ("
            local parenth2 = "\x87)\x80: "
            if value.noset or value.hideoptions or value.seperateoptions then
                parenth1 = ""
                parenth2 = " "
            end


            CONS_Printf(player, "* \x83" .. value.name .."\x80:" .. parenth1 .. options .. parenth2 .. value.description .. setvalue)
        end
    end
end

local function CheckAutoSave(player)
    if player and player.chaoticsettings then
        if player.chaoticsettings.autosave == "on" then
            return true
        end
    else
        local name = "client/chaoticedition/autosave.dat"
        local loadfile = io.openlocal(name, "r")
        if loadfile != "off" then
            return true
        end
    end
end

local function ChaosCommand(player, arg1, arg2, arg3, arg4, arg5, arg6)
    if not arg1 then
        CONS_Printf(player, "\x82No arguments inputted\x80, use \x83\"chaotic help\"\x80 for help.")
        return
    end

    if tonumber(arg2) or arg2 == "0" then
        arg2 = tonumber(arg2)
    end

    if arg1 == "help" then
        PrintHelpForCategory(player, "Options")
        PrintHelpForCategory(player, "HUD")
        PrintHelpForCategory(player, "Tweaks")
        PrintHelpForCategory(player, "Misc.")
        if player.chaotic then 
            player.chaotic.ReadHelp = false
            player.chaoticsettings.ReadHelp = true
            CHT_Save1Setting(player, "ReadHelp")
        end
        -- insert thing here
        return
    elseif arg1 and (not arg2 and arg2 != 0) then
        if GetHelp(arg1) then
            for index, value in ipairs(OptionsTable) do
                if value.name == arg1 and not value.hide then
                    if value.optiontype == "save" then
                        CHT_SaveSettings(player)
                        CONS_Printf(player, "*\x82 Saved settings!")
                        return
                    elseif value.optiontype == "load" then
                        CHT_LoadSettings(player)
                        CONS_Printf(player, "*\x82 Loaded settings!")
                        return
                    end
                    local options = ""
                    local setvalue = ""
                    local function Helper(override)
                        if (not value.hideoptions and not value.seperateoptions) or override then
                            for i2, v2 in ipairs(value.options) do
                                local slash = "\x80/"
                                if value.optiontype == "number" then
                                    slash = "\x80-"
                                end
                                if i2 == tablelength(value.options) then
                                    slash = ""
                                end
                                options = $ .. "\x87" .. v2 .. slash
                            end
                        end
                    end
                    Helper()
                    if not value.hideoptions then
                        setvalue = "\x81 set to: \x83" .. value.set
                        if value.playervar then
                            if player.chaoticsettings then
                                setvalue = "\x81 set to: \x83" .. player.chaoticsettings[value.name]
                            end
                        end
                    end
                    local parenth1 = "\x87 ("
                    local parenth2 = "\x87)\x80: "
                    if value.hideoptions or value.seperateoptions then
                        parenth1 = ""
                        parenth2 = " "
                    end
                    CONS_Printf(player, "* \x83" .. value.Fname .."\x80:" .. parenth1 .. options .. parenth2 .. value.description .. setvalue)
                    if value.seperateoptions then
                        Helper(true)
                        parenth1 = "\x87("
                        parenth2 = "\x87)\x80"
                        CONS_Printf(player, "* \x83" .. parenth1 .. options .. parenth2)
                    end


                    -- print out the table contents here
                    if value.optiontype == "table" then -- and tablelength(value.set) > 0
                        local printer = ""
                        if player.chaoticsettings then
                            for k2, v2 in pairs(player.chaoticsettings[value.name]) do
                                printer = $ .. k2 .. "  "
                            end
                        end
                        CONS_Printf(player, "* \x83" .. "Options set" .."\x80: " .. printer)
                    end
                    if value.nomatch and IsMatchMode() then
                        CONS_Printf(player, "\x83 Notice:\x80 Can not change \x82" .. value.name .. "\x80 while in a match!")
                        return
                    end
                    return
                end
            end
        end
    elseif GetHelp(arg1) then
        for index, value in ipairs(OptionsTable) do
            if value.name == arg1 and (not value.noset or value.optiontype == "table") and not value.hide then
                if value.admin and (netgame and not IsPlayerAdmin(player)) then
                    CONS_Printf(player, "\x85 ERROR:\x80 Can not change \x82" .. value.name .. "\x80 while in a netgame and not an admin!")
                    return
                end
                if value.nomatch and IsMatchMode() then
                    CONS_Printf(player, "\x85 ERROR:\x80 Can not change \x82" .. value.name .. "\x80 while in a match!")
                    return
                end

                if not value.optiontype then
                    for i2, v2 in ipairs(value.options) do -- change option if valid
                        if arg2 == v2 then
                            if not value.playervar then
                                value.set = arg2
                            else
                                if player.chaoticsettings then
                                    player.chaoticsettings[value.name] = arg2
                                else
                                    CONS_Printf(player, "*\x82 NOTICE!\x81 Can not change player variables!\x80 Make sure you're in-game to change settings.")
                                end
                            end
                        end
                    end
                elseif value.optiontype == "number" then
                    if type(arg2) == "number" then
                        if arg2 >= value.options[1] and arg2 <= value.options[2] then -- change option if within range
                            if not value.playervar then
                                value.set = arg2
                            else
                                if player.chaoticsettings then
                                    player.chaoticsettings[value.name] = arg2
                                else
                                    CONS_Printf(player, "*\x82 NOTICE!\x81 Can not change player variables!\x80 Make sure you're in-game to change settings.")
                                end
                            end
                        end
                    end
                elseif value.optiontype == "table" then
                    -- arg1 is for "blacklist"
                    -- arg2 is "add" or "remove"
                    -- arg3 is the actual effect name to be added, i.e. "Earthquake"
                    -- full command should be like this: chaotic blacklist add "Earthquake"
                    if arg2 != "add" and arg2 != "remove" then
                        CONS_Printf(player, "\x82Invalid arguments inputted\x80, use \x83\"add\"\x80 or \x83\"remove\"\x80 to change entries for this setting.")
                        return
                    end

                    local found = false
                    for key2, value2 in pairs(value.options) do
                        print(tostring(key2) .. " " .. tostring(arg3))
                        if arg3 == key2 then
                            if value.playervar then
                                if player.chaoticsettings then
                                    if arg2 == "add" then
                                        player.chaoticsettings[value.name][arg3] = true
                                    elseif arg2 == "remove" then
                                        player.chaoticsettings[value.name][arg3] = nil
                                    end
                                else
                                    CONS_Printf(player, "*\x82 NOTICE!\x81 Can not change player variables!\x80 Make sure you're in-game to change settings.")
                                end
                            else
                                if arg2 == "add" then
                                    value.set[arg3] = true
                                elseif arg2 == "remove" then
                                    value.set[arg3] = nil
                                end
                            end
                            DebugPrint(player, "Option " .. arg3 .. " changed to " .. arg2)
                            found = true
                            break
                        end
                    end
                    if not found then
                        CONS_Printf(player, "\x82Invalid arguments inputted\x80, make sure you capitalized the name correctly, and used quotation marks.")
                        return
                    end
                end

                local setvalue
                local settext = "is set to"
                if not value.optiontype == "table" then
                    setvalue = value.set
                else
                    setvalue = arg3
                    if arg2 == "add" then
                        settext = "has added"
                    elseif arg2 == "remove" then
                        settext = "has removed"
                    end
                end

                if value.playervar and player.chaoticsettings then
                    setvalue = player.chaoticsettings[value.name]
                    if value.optiontype == "table" then
                        setvalue = arg3
                    end
                    if CheckAutoSave(player) then
                        CHT_Save1Setting(player, value.name, (value.optiontype == "table"))
                    end
                    addHUDText(player, "\x81" .. value.name .. "\x80 " .. settext .. " \x82" .. tostring(setvalue))
                end

                CONS_Printf(player, "> \x81" .. value.name .. "\x80 " .. settext .. " \x83" .. tostring(setvalue))
                -- if table, make a seperate print for all the entries here
                return
            end
        end
    end
CONS_Printf(player, "\x82Invalid arguments inputted\x80, use \x83\"chaotic help\"\x80 for help.")
end

COM_AddCommand("chaotic", ChaosCommand)

addHook("NetVars", function(net)
	OptionsTable = net(OptionsTable)
	issomeoneusingblackhole = net(issomeoneusingblackhole)
end)

print("SRB2 Chaotic Edition v" .. SRB2CEversion .." loaded!")